import warnings

import numpy as np
from PIL import Image as PILImage
from mock import patch, MagicMock
from pytest import raises

import menpo.io as mio


def test_import_incorrect_built_in():
    with raises(ValueError):
        mio.import_builtin_asset('adskljasdlkajd.obj')


def test_breaking_bad_import():
    img = mio.import_builtin_asset('breakingbad.jpg')
    assert (img.shape == (1080, 1920))
    assert (img.n_channels == 3)
    assert (img.landmarks['PTS'].n_points == 68)


def test_breaking_bad_import_kwargs():
    img = mio.import_builtin_asset('breakingbad.jpg', normalize=False)
    assert (img.pixels.dtype == np.uint8)


def test_takeo_import():
    img = mio.import_builtin_asset('takeo.ppm')
    assert (img.shape == (225, 150))
    assert (img.n_channels == 3)
    assert (img.landmarks['PTS'].n_points == 68)


def test_einstein_import():
    img = mio.import_builtin_asset('einstein.jpg')
    assert (img.shape == (1024, 817))
    assert (img.n_channels == 1)
    assert (img.landmarks['PTS'].n_points == 68)


def test_lenna_import():
    img = mio.import_builtin_asset('lenna.png')
    assert (img.shape == (512, 512))
    assert (img.n_channels == 3)
    assert (img.landmarks.n_groups == 2)
    assert (img.landmarks['LJSON'].n_points == 68)
    assert (img.landmarks['pupils'].n_points == 2)


def test_import_builtin_ljson():
    lmarks = mio.import_builtin_asset('lenna.ljson')['LJSON']
    assert (lmarks.n_points == 68)


def test_import_builtin_pts():
    lmarks = mio.import_builtin_asset('einstein.pts')['PTS']
    assert (lmarks.n_points == 68)


def test_resolve_from_paths_single_group():
    def resolver(path):
        test_dict = {'test': path.with_name('takeo.pts')}
        return mio.input.resolve_from_paths(test_dict)

    image = mio.import_image(mio.data_path_to('einstein.jpg'),
                             landmark_resolver=resolver)
    assert (image.landmarks.n_groups == 1)
    assert (image.landmarks['test'].path == mio.data_path_to('takeo.pts'))


def test_resolve_from_paths_multi_group():
    def resolver(path):
        test_dict = {'test': path.with_name('lenna.ljson')}
        return mio.input.resolve_from_paths(test_dict)

    image = mio.import_image(mio.data_path_to('einstein.jpg'),
                             landmark_resolver=resolver)
    assert (image.landmarks.n_groups == 2)
    assert (set(image.landmarks.keys()) == {'test_LJSON', 'test_pupils'})


def test_path():
    # choose a random asset (all should have it!)
    img = mio.import_builtin_asset('einstein.jpg')
    path = mio.data_path_to('einstein.jpg')
    assert (img.path == path)
    assert (img.path.stem == 'einstein')
    assert (img.path.suffix == '.jpg')
    assert (img.path.parent == mio.data_dir_path())
    assert (img.path.name == 'einstein.jpg')


@patch('menpo.io.input.base._pathlib_glob_for_pattern')
def test_single_suffix_dot_in_path(pathlib_glob):
    import menpo.io.input.base as mio_base
    from pathlib import Path

    fake_path = Path('fake_path.t0.t1.t2')
    pathlib_glob.return_value = [fake_path]
    ext_map = MagicMock()
    ext_map.__contains__.side_effect = lambda x: x == '.t2'

    ret_val = next(mio_base.glob_with_suffix('*.t0.t1.t2', ext_map))
    assert (ret_val == fake_path)
    ext_map.__contains__.assert_called_with('.t2')


def test_upper_extension_mapped_to_lower():
    import menpo.io.input.base as mio_base
    from pathlib import Path
    ext_map = MagicMock()

    mio_base.importer_for_filepath(Path('fake_path.JPG'), ext_map)
    ext_map.get.assert_called_with('.jpg')


@patch('menpo.io.input.base._pathlib_glob_for_pattern')
def test_double_suffix(pathlib_glob):
    import menpo.io.input.base as mio_base
    from pathlib import Path

    fake_path = Path('fake_path.t1.t2')
    pathlib_glob.return_value = [fake_path]
    ext_map = MagicMock()
    ext_map.__contains__.side_effect = lambda x: x == '.t1.t2'

    ret_val = next(mio_base.glob_with_suffix('*.t1.t2', ext_map))
    assert (ret_val == fake_path)
    ext_map.__contains__.assert_any_call('.t1.t2')
    ext_map.__contains__.assert_any_call('.t2')


def test_import_image():
    img_path = mio.data_dir_path() / 'einstein.jpg'
    im = mio.import_image(img_path)
    assert im.pixels.dtype == np.float
    assert im.n_channels == 1


def test_custom_landmark_resolver():
    def lmark_resolver(path):
        return mio.import_landmark_file(mio.data_path_to('takeo.pts'))

    img = mio.import_image(mio.data_path_to('lenna.png'),
                           landmark_resolver=lmark_resolver)
    assert (img.has_landmarks)

    takeo_lmarks = mio.import_builtin_asset.takeo_pts()['PTS']
    np.allclose(img.landmarks['PTS'].points,
                takeo_lmarks.points)


def test_landmark_resolver_none():
    img = mio.import_image(mio.data_path_to('lenna.png'),
                           landmark_resolver=None)
    assert (not img.has_landmarks)


def test_import_image_no_norm():
    img_path = mio.data_dir_path() / 'einstein.jpg'
    im = mio.import_image(img_path, normalize=False)
    assert im.pixels.dtype == np.uint8


def test_import_landmark_file():
    lm_path = mio.data_dir_path() / 'einstein.pts'
    mio.import_landmark_file(lm_path)


def test_import_images():
    imgs = list(mio.import_images(mio.data_dir_path()))
    imgs_filenames = set(i.path.stem for i in imgs)
    exp_imgs_filenames = {'einstein', 'takeo', 'tongue', 'breakingbad', 'lenna',
                          'menpo_thumbnail'}
    assert exp_imgs_filenames == imgs_filenames


def test_import_images_are_ordered_and_unduplicated():
    # we know that import_images returns images in path order
    imgs = list(mio.import_images(mio.data_dir_path()))
    imgs_filenames = [i.path.stem for i in imgs]
    print(imgs_filenames)
    exp_imgs_filenames = ['breakingbad', 'einstein', 'lenna', 'menpo_thumbnail', 'takeo', 'tongue']
    assert exp_imgs_filenames == imgs_filenames


def test_lsimgs_filenamess():
    assert (set(mio.ls_builtin_assets()) == {'breakingbad.jpg',
                                             'einstein.jpg', 'einstein.pts',
                                             'lenna.png', 'breakingbad.pts',
                                             'lenna.ljson', 'takeo.ppm',
                                             'takeo.pts', 'tongue.jpg',
                                             'tongue.pts',
                                             'menpo_thumbnail.jpg'})


def test_image_paths():
    ls = mio.image_paths(mio.data_dir_path())
    assert (len(list(ls)) == 6)


def test_import_images_wrong_path_raises_value_error():
    with raises(ValueError):
        list(mio.import_images('asldfjalkgjlaknglkajlekjaltknlaekstjlakj'))


def test_import_landmark_files_wrong_path_raises_value_error():
    with raises(ValueError):
        list(mio.import_landmark_files('asldfjalkgjlaknglkajlekjaltknlaekstjlakj'))


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_no_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new('RGBA', (10, 15))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.png', normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 4
    assert im.pixels.dtype == np.uint8


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_RGBA_normalize(is_file, mock_image):
    from menpo.image import MaskedImage

    mock_image.return_value = PILImage.new('RGBA', (10, 15))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.ppm', normalize=True)
    assert im.shape == (15, 10)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.float
    assert type(im) == MaskedImage


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_RGBA_no_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new('RGBA', (10, 15))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.ppm', normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 4
    assert im.pixels.dtype == np.uint8


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_L_no_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new('L', (10, 15))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.ppm', normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.uint8


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_L_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new('L', (10, 15))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.ppm', normalize=True)
    assert im.shape == (15, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.float


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_I_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new('I', (10, 15))
    is_file.return_value = True
    with raises(ValueError):
        mio.import_image('fake_image_being_mocked.ppm', normalize=True)


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_I_no_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new('I', (10, 15))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.ppm', normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.int32


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_1_normalize(is_file, mock_image):
    from menpo.image import BooleanImage

    mock_image.return_value = PILImage.new('1', (10, 15))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.ppm', normalize=True)
    assert im.shape == (15, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.bool
    assert type(im) == BooleanImage


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_1_no_normalize(is_file, mock_image):
    from menpo.image import BooleanImage

    mock_image.return_value = PILImage.new('1', (10, 15))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.ppm', normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.bool
    assert type(im) == BooleanImage


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_1_proper_conversion(is_file, mock_image):
    from menpo.image import BooleanImage

    arr = np.zeros((10, 10), dtype=np.uint8)
    arr[4, 4] = 255
    mock_image.return_value = PILImage.fromarray(arr).convert('1')
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.ppm', normalize=False)
    assert im.shape == (10, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.bool
    assert type(im) == BooleanImage
    assert np.all(im.pixels == arr.astype(np.bool))


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_P_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new('P', (10, 10))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.ppm', normalize=True)
    assert im.shape == (10, 10)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.float


@patch('PIL.Image.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_PIL_P_no_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new('P', (10, 15))
    is_file.return_value = True

    im = mio.import_image('fake_image_being_mocked.ppm', normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.uint8


@patch('subprocess.Popen')
@patch('menpo.io.input.video.video_infos_ffprobe')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_ffmpeg_GIF_normalize(is_file, video_infos_ffprobe, pipe):
    video_infos_ffprobe.return_value = {'duration': 2, 'width': 100,
                                        'height': 150, 'n_frames': 10, 'fps': 5}
    empty_frame = np.zeros(150 * 100 * 3, dtype=np.uint8).tostring()
    pipe.return_value.stdout.read.return_value = empty_frame
    is_file.return_value = True

    ll = mio.import_image('fake_image_being_mocked.gif', normalize=True)
    assert ll.path.name == 'fake_image_being_mocked.gif'
    assert ll.fps == 5
    assert len(ll) == 10

    im = ll[0]
    assert im.shape == (150, 100)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.float


@patch('subprocess.Popen')
@patch('menpo.io.input.video.video_infos_ffprobe')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_ffmpeg_GIF_no_normalize(is_file, video_infos_ffprobe, pipe):
    video_infos_ffprobe.return_value = {'duration': 2, 'width': 100,
                                        'height': 150, 'n_frames': 10, 'fps': 5}
    empty_frame = np.zeros(150 * 100 * 3, dtype=np.uint8).tostring()
    pipe.return_value.stdout.read.return_value = empty_frame
    is_file.return_value = True

    ll = mio.import_image('fake_image_being_mocked.gif', normalize=False)
    assert ll.path.name == 'fake_image_being_mocked.gif'
    assert ll.fps == 5
    assert len(ll) == 10

    im = ll[0]
    assert im.shape == (150, 100)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.uint8


@patch('menpo.io.input.landmark.json.load')
@patch('menpo.io.input.base.Path.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_v1_ljson_null_values(is_file, mock_open, mock_dict):
    v1_ljson = {"groups": [
        {"connectivity": [[0, 1], [1, 2], [2, 3]],
         "label": "chin", "landmarks": [
            {"point": [987.9, 1294.1]}, {"point": [96.78, 1246.8]},
            {"point": [None, 0.1]}, {"point": [303.22, 167.2]}]},
        {"connectivity": [[0, 1]],
         "label": "leye", "landmarks": [
            {"point": [None, None]},
            {"point": [None, None]}]}
    ], "version": 1}
    mock_dict.return_value = v1_ljson
    is_file.return_value = True

    with warnings.catch_warnings(record=True) as found_warnings:
        lmark = mio.import_landmark_file('fake_lmark_being_mocked.ljson',
                                         group='LJSON')
    nan_points = np.isnan(lmark.points)

    # Should raise deprecation warning
    assert len(found_warnings) == 1, [w.message for w in found_warnings]
    assert nan_points[2, 0]  # y-coord None point is nan
    assert not nan_points[2, 1]  # x-coord point is not nan
    assert np.all(nan_points[4:, :])  # all of leye label is nan


@patch('menpo.io.input.landmark.json.load')
@patch('menpo.io.input.base.Path.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_v2_ljson_null_values(is_file, mock_open, mock_dict):
    v2_ljson = {"labels": [
        {"label": "left_eye", "mask": [0, 1, 2]},
        {"label": "right_eye", "mask": [3, 4, 5]}
    ],
        "landmarks": {
            "connectivity": [[0, 1], [1, 2], [2, 0], [3, 4],
                             [4, 5], [5, 3]],
            "points": [[None, 200.5], [None, None],
                       [316.8, 199.15], [339.48, 205.0],
                       [358.54, 217.82], [375.0, 233.4]]
        },
        "version": 2}

    mock_dict.return_value = v2_ljson
    is_file.return_value = True
    with warnings.catch_warnings(record=True) as w:
        lmark = mio.import_landmark_file('fake_lmark_being_mocked.ljson',
                                         group='LJSON')
    nan_points = np.isnan(lmark.points)
    assert nan_points[0, 0]  # y-coord None point is nan
    assert not nan_points[0, 1]  # x-coord point is not nan
    assert np.all(nan_points[1, :])  # all of leye label is nan


@patch('menpo.io.input.landmark.json.load')
@patch('menpo.io.input.base.Path.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_v3_ljson_null_values(is_file, mock_open, mock_dict):
    v3_ljson = {
        "groups": {
            "LJSON": {
                "labels": [
                    {"label": "left_eye", "mask": [0, 1, 2]},
                    {"label": "right_eye", "mask": [3, 4, 5]}
                ],
                "landmarks": {
                    "connectivity": [[0, 1], [1, 2], [2, 0], [3, 4],
                                     [4, 5], [5, 3]],
                    "points": [[None, 200.5], [None, None],
                               [316.8, 199.15], [339.48, 205.0],
                               [358.54, 217.82], [375.0, 233.4]]
                }
            }
        },
        "version": 3
    }

    mock_dict.return_value = v3_ljson
    is_file.return_value = True
    lmark_dict = mio.import_landmark_file('fake_lmark_being_mocked.ljson')
    assert isinstance(lmark_dict, dict)
    lmark = lmark_dict['LJSON']
    nan_points = np.isnan(lmark.points)
    assert nan_points[0, 0]  # y-coord None point is nan
    assert not nan_points[0, 1]  # x-coord point is not nan
    assert np.all(nan_points[1, :])  # all of leye label is nan


@patch('random.shuffle')
def test_shuffle_kwarg_true_calls_shuffle(mock):
    list(mio.import_images(mio.data_dir_path(), shuffle=True))
    assert mock.called


def test_import_as_generator():
    import types
    gen = mio.import_images(mio.data_dir_path(), as_generator=True)
    assert isinstance(gen, types.GeneratorType)
    gen = mio.import_landmark_files(mio.data_dir_path(), as_generator=True)
    assert isinstance(gen, types.GeneratorType)


def test_import_lazy_list():
    from menpo.base import LazyList
    data_path = mio.data_dir_path()
    ll = mio.import_images(data_path)
    assert isinstance(ll, LazyList)
    ll = mio.import_landmark_files(data_path)
    assert isinstance(ll, LazyList)


@patch('menpo.io.input.pickle.pickle.load')
@patch('menpo.io.input.base.Path.open')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_pickle(is_file, mock_open, mock_pickle):
    mock_pickle.return_value = {'test': 1}
    is_file.return_value = True

    objs = mio.import_pickle('mocked.pkl')
    assert isinstance(objs, dict)
    assert 'test' in objs
    assert objs['test'] == 1


@patch('menpo.io.input.pickle.pickle.load')
@patch('menpo.io.input.base.Path.open')
@patch('menpo.io.input.base.Path.is_file')
@patch('sys.version_info')
def test_importing_pickle_encoding_py3(version_info, is_file, mock_open,
                                       mock_pickle):
    version_info.major = 3
    mock_pickle.return_value = {'test': 1}
    is_file.return_value = True

    mio.import_pickle('mocked.pkl', encoding='latin1')
    assert mock_pickle.call_args[1].get('encoding') == 'latin1'


@patch('menpo.io.input.pickle.pickle.load')
@patch('menpo.io.input.base.Path.open')
@patch('menpo.io.input.base.Path.is_file')
@patch('sys.version_info')
def test_importing_pickle_encoding_ignored_py2(version_info, is_file, mock_open,
                                               mock_pickle):
    version_info.major = 2
    mock_pickle.return_value = {'test': 1}
    is_file.return_value = True

    mio.import_pickle('mocked.pkl', encoding='latin1')
    assert 'encoding' not in mock_pickle.call_args[1]


@patch('menpo.io.input.pickle.pickle.load')
@patch('menpo.io.input.base.Path.open')
@patch('menpo.io.input.base.Path.glob')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_pickles(is_file, glob, mock_open, mock_pickle):
    from pathlib import Path
    from menpo.base import LazyList
    mock_pickle.return_value = {'test': 1}
    is_file.return_value = True
    glob.return_value = [Path('mocked1.pkl'), Path('mocked2.pkl')]

    objs = mio.import_pickles('*')
    assert isinstance(objs, LazyList)
    assert len(objs) == 2
    assert objs[0]['test'] == 1
    assert objs[1]['test'] == 1


@patch('menpo.io.input.pickle.pickle.load')
@patch('menpo.io.input.base.Path.open')
@patch('menpo.io.input.base.Path.glob')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_pickles_as_generator(is_file, glob, mock_open, mock_pickle):
    from pathlib import Path
    import types
    mock_pickle.return_value = {'test': 1}
    is_file.return_value = True
    glob.return_value = [Path('mocked1.pkl'), Path('mocked2.pkl')]

    objs = mio.import_pickles('*', as_generator=True)
    assert isinstance(objs, types.GeneratorType)
    objs = list(objs)
    assert len(objs) == 2
    assert objs[0]['test'] == 1
    assert objs[1]['test'] == 1


@patch('subprocess.Popen')
@patch('menpo.io.input.video.video_infos_ffprobe')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_ffmpeg_avi_no_normalize(is_file, video_infos_ffprobe, pipe):
    video_infos_ffprobe.return_value = {'duration': 2, 'width': 100,
                                        'height': 150, 'n_frames': 10, 'fps': 5}
    is_file.return_value = True
    empty_frame = np.zeros(150 * 100 * 3, dtype=np.uint8).tostring()
    pipe.return_value.stdout.read.return_value = empty_frame
    ll = mio.import_video('fake_image_being_mocked.avi', normalize=False)
    assert ll.path.name == 'fake_image_being_mocked.avi'
    assert ll.fps == 5
    assert len(ll) == 5 * 2
    image = ll[0]
    assert image.shape == ((150, 100))
    assert image.n_channels == 3
    assert image.pixels.dtype == np.uint8


@patch('subprocess.Popen')
@patch('menpo.io.input.video.video_infos_ffprobe')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_ffmpeg_avi_normalize(is_file, video_infos_ffprobe, pipe):
    video_infos_ffprobe.return_value = {'duration': 2, 'width': 100,
                                        'height': 150, 'n_frames': 10, 'fps': 5}
    is_file.return_value = True
    empty_frame = np.zeros(150 * 100 * 3, dtype=np.uint8).tostring()
    pipe.return_value.stdout.read.return_value = empty_frame
    ll = mio.import_video('fake_image_being_mocked.avi', normalize=True)
    assert ll.path.name == 'fake_image_being_mocked.avi'
    assert ll.fps == 5
    assert len(ll) == 5 * 2
    image = ll[0]
    assert image.shape == (150, 100)
    assert image.n_channels == 3
    assert image.pixels.dtype == np.float


@patch('menpo.io.input.video.video_infos_ffprobe')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_ffmpeg_exact_frame_count_no_ffprobe(is_file, video_infos_ffprobe):
    video_infos_ffprobe.side_effect = ValueError
    is_file.return_value = True
    with raises(ValueError):
        mio.import_video('fake_image_being_mocked.avi', normalize=True,
                         exact_frame_count=True)


@patch('subprocess.Popen')
@patch('menpo.io.input.video.video_infos_ffprobe')
@patch('menpo.io.input.video.video_infos_ffmpeg')
@patch('menpo.io.input.base.Path.is_file')
def test_importing_ffmpeg_no_exact_frame_count_no_ffprobe(
        is_file, video_infos_ffmpeg, video_infos_ffprobe, pipe):
    video_infos_ffprobe.side_effect = ValueError
    video_infos_ffmpeg.return_value = {'duration': 2, 'width': 100,
                                       'height': 150, 'n_frames': 10, 'fps': 5}
    is_file.return_value = True
    empty_frame = np.zeros(150 * 100 * 3, dtype=np.uint8).tostring()
    pipe.return_value.stdout.read.return_value = empty_frame
    ll = mio.import_video('fake_image_being_mocked.avi', normalize=True,
                          exact_frame_count=False)
    assert ll.path.name == 'fake_image_being_mocked.avi'
    assert ll.fps == 5
    assert len(ll) == 5 * 2
    image = ll[0]
    assert image.shape == (150, 100)
    assert image.n_channels == 3
    assert image.pixels.dtype == np.float


def test_import_images_negative_max_images():
    with raises(ValueError):
        list(mio.import_images(mio.data_dir_path(), max_images=-2))


def test_import_images_zero_max_images():
    with raises(ValueError):
        # different since the conditional 'if max_assets' is skipped,
        # thus all images might be imported.
        list(mio.import_images(mio.data_dir_path(), max_images=0))


# TODO: remove once the normalise argument is removed.
def test_import_image_deprecated_normalise_kwarg():
    with warnings.catch_warnings(record=True) as w:
        img = mio.import_builtin_asset('breakingbad.jpg', normalise=False)
        assert len(w) == 1
    assert img.pixels.dtype == np.uint8


@patch('menpo.io.input.base.Path.is_file')
def test_register_image_importer(is_file):
    from menpo.image import Image
    image = Image.init_blank((10, 10))

    def foo_importer(filepath, **kwargs):
        return image

    is_file.return_value = True

    with patch.dict(mio.input.extensions.image_types, {}, clear=True):
        mio.register_image_importer('.foo', foo_importer)
        new_image = mio.import_image('fake.foo')
    assert image is new_image


@patch('menpo.io.input.base.Path.is_file')
def test_register_landmark_importer(is_file):
    from menpo.shape import PointCloud
    lmark = PointCloud.init_2d_grid((1, 1))

    def foo_importer(filepath, **kwargs):
        return lmark

    is_file.return_value = True

    with patch.dict(mio.input.extensions.image_landmark_types, {}, clear=True):
        mio.register_landmark_importer('.foo', foo_importer)
        new_lmark = mio.import_landmark_file('fake.foo')
    assert lmark is new_lmark


@patch('menpo.io.input.base.Path.is_file')
def test_register_video_importer(is_file):
    from menpo.image import Image
    from menpo.base import LazyList

    def foo_importer(filepath, **kwargs):
        return LazyList([lambda: Image.init_blank((10, 10))])

    is_file.return_value = True

    with patch.dict(mio.input.extensions.ffmpeg_video_types, {}, clear=True):
        mio.register_video_importer('.foo', foo_importer)
        new_video = mio.import_video('fake.foo')
    assert len(new_video) == 1


@patch('menpo.io.input.base.Path.is_file')
def test_register_pickle_importer(is_file):
    obj = object()

    def foo_importer(filepath, **kwargs):
        return obj

    is_file.return_value = True

    with patch.dict(mio.input.extensions.pickle_types, {}, clear=True):
        mio.register_pickle_importer('.foo', foo_importer)
        new_obj = mio.import_pickle('fake.foo')
    assert new_obj is obj


def test_register_no_leading_period():
    ext_map = {}
    mio.input.base._register_importer(ext_map, 'foo', lambda x: x)
    assert '.foo' in ext_map
    assert 'foo' not in ext_map



from mock import Mock
from pytest import raises
import numpy as np
from numpy.testing import assert_equal

from menpo.transform.base import Alignment


# Mock a pointcloud
source_array = np.random.random([5, 2])
target_array = np.random.random([5, 2])
target_mismatch_array = np.random.random([8, 2])
target_array_3d = np.random.random([5, 3])
source = Mock(points=source_array, n_dims=source_array.shape[1],
              n_points=source_array.shape[0])
target = Mock(points=target_array, n_dims=target_array.shape[1],
              n_points=target_array.shape[0])
target_mismatch = Mock(points=target_mismatch_array,
                       n_dims=target_mismatch_array.shape[1],
                       n_points=target_mismatch_array.shape[0])
target_3d = Mock(points=target_array_3d, n_dims=target_array_3d.shape[1],
                 n_points=target_array_3d.shape[0])


class MockedAlignment(Alignment):
    def _sync_state_from_target(self):
        pass

    def apply(self, x):
        return x


def test_alignment_init():
    MockedAlignment(source, target)


def test_alignment_init_n_dims_mismatch():
    with raises(ValueError):
        MockedAlignment(source, target_3d)


def test_alignment_init_n_dims_mismatch():
    with raises(ValueError):
        MockedAlignment(source, target_3d)


def test_alignment_init_n_points_mismatch():
    with raises(ValueError):
        MockedAlignment(source, target_mismatch)


def test_alignment_source():
    al = MockedAlignment(source, target)
    assert (al.source is source)


def test_alignment_target():
    al = MockedAlignment(source, target)
    assert (al.target is target)


def test_alignment_aligned_source():
    al = MockedAlignment(source, target)
    assert (al.aligned_source() is source)


def test_alignment_alignment_error():
    al = MockedAlignment(source, target)
    error = al.alignment_error()
    assert_equal(error, np.linalg.norm(al.source.points - al.target.points))


def test_alignment_view_non_2d():
    al = MockedAlignment(target_3d, target_3d)
    with raises(NotImplementedError):
        al.view()



import numpy as np
import scipy.sparse as sparse

from . import (PointCloud, UndirectedGraph, DirectedGraph, Tree, TriMesh,
               PointUndirectedGraph, PointDirectedGraph, PointTree)


def stencil_grid(stencil, shape, dtype=None, format=None):
    """Construct a sparse matrix form a local matrix stencil

    This function is useful for building sparse adjacency matrices according
    to a specific connectivity pattern.

    This function is borrowed from the PyAMG project, under the permission of
    the MIT license:

    The MIT License (MIT)

    Copyright (c) 2008-2015 PyAMG Developers

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.

    The original version of this file can be found here:

    https://github.com/pyamg/pyamg/blob/621d63411895898660e5ea078840118905bec061/pyamg/gallery/stencil.py

    This file has been modified to fit the style standards of the Menpo
    project.

    Parameters
    ----------
    S : `ndarray`
        Matrix stencil stored in N-d array
    grid : `tuple`
        Tuple containing the N shape dimensions (shape)
    dtype : `np.dtype`, optional
        Numpy data type of the result
    format : `str`, optional
        Sparse matrix format to return, e.g. "csr", "coo", etc.

    Returns
    -------
    A : sparse matrix
        Sparse matrix which represents the operator given by applying
        stencil stencil at each vertex of a regular shape with given dimensions.

    Notes
    -----
    The shape vertices are enumerated as ``arange(prod(shape)).reshape(shape)``.
    This implies that the last shape dimension cycles fastest, while the
    first dimension cycles slowest.  For example, if ``shape=(2,3)`` then the
    shape vertices are ordered as ``(0,0), (0,1), (0,2), (1,0), (1,1), (1,2)``.

    This coincides with the ordering used by the NumPy functions
    ``ndenumerate()`` and ``mgrid()``.

    Raises
    ------
    ValueError
        If the stencil shape is not odd.
    ValueError
        If the stencil dimension does not equal the number of shape dimensions
    ValueError
        If the shape dimensions are not all positive

    Examples
    --------
    >>> import numpy as np
    >>> from menpo.shape import stencil_grid
    >>> stencil = [[0,-1,0],[-1,4,-1],[0,-1,0]]  # 2D Poisson stencil
    >>> shape = (3, 3)                           # 2D shape with shape 3x3
    >>> A = stencil_grid(stencil, shape, dtype=np.float, format='csr')
    >>> A.todense()
    matrix([[ 4., -1.,  0., -1.,  0.,  0.,  0.,  0.,  0.],
            [-1.,  4., -1.,  0., -1.,  0.,  0.,  0.,  0.],
            [ 0., -1.,  4.,  0.,  0., -1.,  0.,  0.,  0.],
            [-1.,  0.,  0.,  4., -1.,  0., -1.,  0.,  0.],
            [ 0., -1.,  0., -1.,  4., -1.,  0., -1.,  0.],
            [ 0.,  0., -1.,  0., -1.,  4.,  0.,  0., -1.],
            [ 0.,  0.,  0., -1.,  0.,  0.,  4., -1.,  0.],
            [ 0.,  0.,  0.,  0., -1.,  0., -1.,  4., -1.],
            [ 0.,  0.,  0.,  0.,  0., -1.,  0., -1.,  4.]])

    >>> stencil = [[0,1,0],[1,0,1],[0,1,0]]  # 2D Lattice Connectivity
    >>> shape = (3, 3)                       # 2D shape with shape 3x3
    >>> A = stencil_grid(stencil, shape, dtype=np.float, format='csr')
    >>> A.todense()
    matrix([[ 0.,  1.,  0.,  1.,  0.,  0.,  0.,  0.,  0.],
            [ 1.,  0.,  1.,  0.,  1.,  0.,  0.,  0.,  0.],
            [ 0.,  1.,  0.,  0.,  0.,  1.,  0.,  0.,  0.],
            [ 1.,  0.,  0.,  0.,  1.,  0.,  1.,  0.,  0.],
            [ 0.,  1.,  0.,  1.,  0.,  1.,  0.,  1.,  0.],
            [ 0.,  0.,  1.,  0.,  1.,  0.,  0.,  0.,  1.],
            [ 0.,  0.,  0.,  1.,  0.,  0.,  0.,  1.,  0.],
            [ 0.,  0.,  0.,  0.,  1.,  0.,  1.,  0.,  1.],
            [ 0.,  0.,  0.,  0.,  0.,  1.,  0.,  1.,  0.]])

    """
    stencil = np.asarray(stencil, dtype=dtype)
    shape = tuple(shape)

    if not (np.asarray(stencil.shape) % 2 == 1).all():
        raise ValueError('all stencil dimensions must be odd')

    if len(shape) != np.ndim(stencil):
        raise ValueError('stencil dimension must equal number of shape\
                          dimensions')

    if min(shape) < 1:
        raise ValueError('shape dimensions must be positive')

    N_v = np.prod(shape)        # number of vertices in the mesh
    N_s = (stencil != 0).sum()  # number of nonzero stencil entries

    # diagonal offsets
    diags = np.zeros(N_s, dtype=int)

    # compute index offset of each dof within the stencil
    strides = np.cumprod([1] + list(reversed(shape)))[:-1]
    indices = tuple(i.copy() for i in stencil.nonzero())
    for i, s in zip(indices, stencil.shape):
        i -= s // 2
        # i = (i - s) // 2
        # i = i // 2
        # i = i - (s // 2)
    for stride, coords in zip(strides, reversed(indices)):
        diags += stride * coords

    data = stencil[stencil != 0].repeat(N_v).reshape(N_s, N_v)

    indices = np.vstack(indices).T

    # zero boundary connections
    for index, diag in zip(indices, data):
        diag = diag.reshape(shape)
        for n, i in enumerate(index):
            if i > 0:
                s = [slice(None)] * len(shape)
                s[n] = slice(0, i)
                diag[tuple(s)] = 0
            elif i < 0:
                s = [slice(None)]*len(shape)
                s[n] = slice(i, None)
                diag[tuple(s)] = 0

    # remove diagonals that lie outside matrix
    mask = abs(diags) < N_v
    if not mask.all():
        diags = diags[mask]
        data = data[mask]

    # sum duplicate diagonals
    if len(np.unique(diags)) != len(diags):
        new_diags = np.unique(diags)
        new_data = np.zeros((len(new_diags), data.shape[1]),
                            dtype=data.dtype)

        for dia, dat in zip(diags, data):
            n = np.searchsorted(new_diags, dia)
            new_data[n, :] += dat

        diags = new_diags
        data = new_data

    return sparse.dia_matrix((data, diags),
                             shape=(N_v, N_v)).asformat(format)


def _get_points_and_number_of_vertices(shape):
    if isinstance(shape, PointCloud):
        return shape.points, shape.n_points
    else:
        raise ValueError("shape must be PointCloud instance.")


def _get_star_graph_edges(vertices_list, root_vertex):
    edges = []
    for v in vertices_list:
        if v != root_vertex:
            edges.append([root_vertex, v])
    return edges


def _get_complete_graph_edges(vertices_list):
    n_vertices = len(vertices_list)
    edges = []
    for i in range(n_vertices-1):
        k = i + 1
        for j in range(k, n_vertices, 1):
            v1 = vertices_list[i]
            v2 = vertices_list[j]
            edges.append([v1, v2])
    return edges


def _get_chain_graph_edges(vertices_list, closed):
    n_vertices = len(vertices_list)
    edges = []
    for i in range(n_vertices-1):
        k = i + 1
        v1 = vertices_list[i]
        v2 = vertices_list[k]
        edges.append([v1, v2])
    if closed:
        v1 = vertices_list[-1]
        v2 = vertices_list[0]
        edges.append([v1, v2])
    return edges


def empty_graph(shape, return_pointgraph=True):
    r"""
    Returns an empty graph given the landmarks configuration of a shape
    instance.

    Parameters
    ----------
    shape : :map:`PointCloud` or subclass
        The shape instance that defines the landmarks configuration based on
        which the graph will be created.
    return_pointgraph : `bool`, optional
        If ``True``, then a :map:`PointUndirectedGraph` instance will be
        returned. If ``False``, then an :map:`UndirectedGraph` instance will be
        returned.

    Returns
    -------
    graph : :map:`UndirectedGraph` or :map:`PointUndirectedGraph`
        The generated graph.
    """
    # get points and number of vertices
    points, n_vertices = _get_points_and_number_of_vertices(shape)

    # create empty edges
    edges = None

    # return graph
    if return_pointgraph:
        return PointUndirectedGraph.init_from_edges(points, edges, n_vertices,
                                                    skip_checks=True)
    else:
        return UndirectedGraph.init_from_edges(edges, n_vertices,
                                               skip_checks=True)


def star_graph(shape, root_vertex, graph_cls=PointTree):
    r"""
    Returns a star graph given the landmarks configuration of a shape instance.

    Parameters
    ----------
    shape : :map:`PointCloud` or subclass
        The shape instance that defines the landmarks configuration based on
        which the graph will be created.
    root_vertex : `int`
        The root of the star tree.
    graph_cls : `Graph` or `PointGraph` subclass
        The output graph type.
        Possible options are ::

            {:map:`UndirectedGraph`, :map:`DirectedGraph`, :map:`Tree`,
             :map:`PointUndirectedGraph`, :map:`PointDirectedGraph`,
             :map:`PointTree`}

    Returns
    -------
    graph : `Graph` or `PointGraph` subclass
        The generated graph.

    Raises
    ------
    ValueError
        graph_cls must be UndirectedGraph, DirectedGraph, Tree,
        PointUndirectedGraph, PointDirectedGraph or PointTree.
    """
    # get points and number of vertices
    points, n_vertices = _get_points_and_number_of_vertices(shape)

    # create star graph edges
    edges = _get_star_graph_edges(range(n_vertices), root_vertex)

    # return graph
    if graph_cls == Tree:
        return graph_cls.init_from_edges(edges=edges, n_vertices=n_vertices,
                                         root_vertex=root_vertex,
                                         skip_checks=True)
    elif graph_cls == PointTree:
        return graph_cls.init_from_edges(points=points, edges=edges,
                                         root_vertex=root_vertex,
                                         skip_checks=True)
    elif graph_cls == UndirectedGraph or graph_cls == DirectedGraph:
        return graph_cls.init_from_edges(edges=edges, n_vertices=n_vertices,
                                         skip_checks=True)
    elif graph_cls == PointUndirectedGraph or graph_cls == PointDirectedGraph:
        return graph_cls.init_from_edges(points=points, edges=edges,
                                         skip_checks=True)
    else:
        raise ValueError("graph_cls must be UndirectedGraph, DirectedGraph, "
                         "Tree, PointUndirectedGraph, PointDirectedGraph or "
                         "PointTree.")


def complete_graph(shape, graph_cls=PointUndirectedGraph):
    r"""
    Returns a complete graph given the landmarks configuration of a shape
    instance.

    Parameters
    ----------
    shape : :map:`PointCloud` or subclass
        The shape instance that defines the landmarks configuration based on
        which the graph will be created.
    graph_cls : `Graph` or `PointGraph` subclass
        The output graph type.
        Possible options are ::

            {:map:`UndirectedGraph`, :map:`DirectedGraph`,
             :map:`PointUndirectedGraph`, :map:`PointDirectedGraph`}

    Returns
    -------
    graph : `Graph` or `PointGraph` subclass
        The generated graph.

    Raises
    ------
    ValueError
        graph_cls must be UndirectedGraph, DirectedGraph, PointUndirectedGraph
        or PointDirectedGraph.
    """
    # get points and number of vertices
    points, n_vertices = _get_points_and_number_of_vertices(shape)

    # create complete graph edges
    edges = _get_complete_graph_edges(range(n_vertices))

    # return graph
    if graph_cls == UndirectedGraph or graph_cls == DirectedGraph:
        return graph_cls.init_from_edges(edges=edges, n_vertices=n_vertices,
                                         skip_checks=True)
    elif graph_cls == PointUndirectedGraph or graph_cls == PointDirectedGraph:
        return graph_cls.init_from_edges(points=points, edges=edges,
                                         skip_checks=True)
    else:
        raise ValueError("graph_cls must be UndirectedGraph, DirectedGraph, "
                         "PointUndirectedGraph or PointDirectedGraph.")


def chain_graph(shape, graph_cls=PointDirectedGraph, closed=False):
    r"""
    Returns a chain graph given the landmarks configuration of a shape instance.

    Parameters
    ----------
    shape : :map:`PointCloud` or subclass
        The shape instance that defines the landmarks configuration based on
        which the graph will be created.
    graph_cls : `Graph` or `PointGraph` subclass
        The output graph type.
        Possible options are ::

            {:map:`UndirectedGraph`, :map:`DirectedGraph`, :map:`Tree`,
             :map:`PointUndirectedGraph`, :map:`PointDirectedGraph`,
             :map:`PointTree`}

    closed : `bool`, optional
        If ``True``, then the chain will be closed (i.e. edge between the
        first and last vertices).

    Returns
    -------
    graph : `Graph` or `PointGraph` subclass
        The generated graph.

    Raises
    ------
    ValueError
        A closed chain graph cannot be a Tree or PointTree instance.
    ValueError
        graph_cls must be UndirectedGraph, DirectedGraph, Tree,
        PointUndirectedGraph, PointDirectedGraph or PointTree.
    """
    # get points and number of vertices
    points, n_vertices = _get_points_and_number_of_vertices(shape)

    # create chain graph edges
    edges = _get_chain_graph_edges(range(n_vertices), closed=closed)

    # return graph
    if graph_cls == Tree:
        if closed:
            raise ValueError("A closed chain graph cannot be a Tree "
                             "instance.")
        else:
            return graph_cls.init_from_edges(edges=edges, n_vertices=n_vertices,
                                             root_vertex=0, skip_checks=True)
    elif graph_cls == PointTree:
        if closed:
            raise ValueError("A closed chain graph cannot be a PointTree "
                             "instance.")
        else:
            return graph_cls.init_from_edges(points=points, edges=edges,
                                             root_vertex=0, skip_checks=True)
    elif graph_cls == UndirectedGraph or graph_cls == DirectedGraph:
        return graph_cls.init_from_edges(edges=edges, n_vertices=n_vertices,
                                         skip_checks=True)
    elif graph_cls == PointUndirectedGraph or graph_cls == PointDirectedGraph:
        return graph_cls.init_from_edges(points=points, edges=edges,
                                         skip_checks=True)
    else:
        raise ValueError("graph_cls must be UndirectedGraph, DirectedGraph, "
                         "Tree, PointUndirectedGraph, PointDirectedGraph or "
                         "PointTree.")


def delaunay_graph(shape, return_pointgraph=True):
    r"""
    Returns a graph with the edges being generated by Delaunay triangulation.

    Parameters
    ----------
    shape : :map:`PointCloud` or subclass
        The shape instance that defines the landmarks configuration based on
        which the graph will be created.
    return_pointgraph : `bool`, optional
        If ``True``, then a :map:`PointUndirectedGraph` instance will be
        returned. If ``False``, then an :map:`UndirectedGraph` instance will be
        returned.

    Returns
    -------
    graph : :map:`UndirectedGraph` or :map:`PointUndirectedGraph`
        The generated graph.
    """
    # get TriMesh instance that estimates the Delaunay triangulation
    if isinstance(shape, PointCloud):
        trimesh = TriMesh(shape.points)
        n_vertices = shape.n_points
        points = shape.points
    else:
        raise ValueError("shape must be a PointCloud instance or subclass.")

    # get edges
    edges = trimesh.edge_indices()

    # return graph
    if return_pointgraph:
        return PointUndirectedGraph.init_from_edges(
            points=points, edges=edges, skip_checks=True)
    else:
        return UndirectedGraph.init_from_edges(
            edges=edges, n_vertices=n_vertices, skip_checks=True)



import numpy as np


def mask_adjacency_array(mask, adjacency_array):
    # Find the indices that have been asked to be removed
    indices_to_remove = np.nonzero(~mask)[0]
    # Set intersection to find any rows containing those elements,
    # reshape back in to the same size as adjacency array
    entries_to_remove = np.in1d(adjacency_array, indices_to_remove)
    entries_to_remove = entries_to_remove.reshape([-1,
                                                   adjacency_array.shape[1]])
    # Only keep those entries that are not flagged for removal
    indices_to_keep = ~entries_to_remove.any(axis=1)
    return adjacency_array[indices_to_keep, :]


def reindex_adjacency_array(adjacency_array):
    remap_vector = np.arange(np.max(adjacency_array) + 1)
    unique_values = np.unique(adjacency_array)
    remap_vector[unique_values] = np.arange(unique_values.shape[0])

    # Apply the mask
    return remap_vector[adjacency_array]



import warnings
import numpy as np

from menpo.base import Copyable, MenpoDeprecationWarning


class Transform(Copyable):
    r"""
    Abstract representation of any spatial transform.

    Provides a unified interface to apply the transform with
    :meth:`apply_inplace` and :meth:`apply`.

    All Transforms support basic composition to form a :map:`TransformChain`.

    There are two useful forms of composition. Firstly, the mathematical
    composition symbol `o` has the following definition::

        Let a(x) and b(x) be two transforms on x.
        (a o b)(x) == a(b(x))

    This functionality is provided by the :meth:`compose_after` family of
    methods: ::

        (a.compose_after(b)).apply(x) == a.apply(b.apply(x))

    Equally useful is an inversion the order of composition - so that over
    time a large chain of transforms can be built to do a useful job, and
    composing on this chain adds another transform to the end (after all other
    preceding transforms have been performed).

    For instance, let's say we want to rescale a :map:`PointCloud` ``p`` around
    its mean, and then translate it some place else. It would be nice to be able
    to do something like::

        t = Translation(-p.centre)  # translate to centre
        s = Scale(2.0)  # rescale
        move = Translate([10, 0 ,0])  # budge along the x axis
        t.compose(s).compose(-t).compose(move)

    In Menpo, this functionality is provided by the :meth:`compose_before()`
    family of methods::

        (a.compose_before(b)).apply(x) == b.apply(a.apply(x))

    For native composition, see the :map:`ComposableTransform` subclass and
    the :map:`VComposable` mix-in.

    For inversion, see the :map:`Invertible` and :map:`VInvertible` mix-ins.

    For alignment, see the :map:`Alignment` mix-in.
    """

    @property
    def n_dims(self):
        r"""
        The dimensionality of the data the transform operates on.

        ``None`` if the transform is not dimension specific.

        :type: `int` or ``None``
        """
        return None

    @property
    def n_dims_output(self):
        r"""
        The output of the data from the transform.

        ``None`` if the output of the transform is not dimension specific.

        :type: `int` or ``None``
        """
        # most Transforms don't change the dimensionality of their input.
        return self.n_dims

    def _apply(self, x, **kwargs):
        r"""
        Applies the transform to the array ``x``, returning the result.

        This method does the actual work of transforming the data, and is the
        one that subclasses must implement. :meth:`apply` and
        :meth:`apply_inplace` both call this method to do that actual work.

        Parameters
        ----------
        x : ``(n_points, n_dims)`` `ndarray`
            The array to be transformed.
        kwargs : `dict`
            Subclasses may need these in their ``_apply`` methods.

        Returns
        -------
        transformed : ``(n_points, n_dims_output)`` `ndarray`
            The transformed array
        """
        raise NotImplementedError()

    def apply_inplace(self, *args, **kwargs):
        r"""
        Deprecated as public supported API, use the non-mutating `apply()`
        instead.

        For internal performance-specific uses, see `_apply_inplace()`.

        """
        warnings.warn('the public API for inplace operations is deprecated '
                      'and will be removed in a future version of Menpo. '
                      'Use .apply() instead.', MenpoDeprecationWarning)
        return self._apply_inplace(*args, **kwargs)

    def _apply_inplace(self, x, **kwargs):
        r"""
        Applies this transform to a :map:`Transformable` ``x`` destructively.

        Any ``kwargs`` will be passed to the specific transform :meth:`_apply`
        method.

        Note that this is an inplace operation that should be used sparingly,
        by internal API's where creating a copy of the transformed object is
        expensive. It does not return anything, as the operation is inplace.

        Parameters
        ----------
        x : :map:`Transformable`
            The :map:`Transformable` object to be transformed.
        kwargs : `dict`
            Passed through to :meth:`_apply`.
        """

        def transform(x_):
            """
            Local closure which calls the :meth:`_apply` method with the
            `kwargs` attached.
            """
            return self._apply(x_, **kwargs)

        try:
            x._transform_inplace(transform)
        except AttributeError:
            raise ValueError('apply_inplace can only be used on Transformable'
                             ' objects.')

    def apply(self, x, batch_size=None, **kwargs):
        r"""
        Applies this transform to ``x``.

        If ``x`` is :map:`Transformable`, ``x`` will be handed this transform
        object to transform itself non-destructively (a transformed copy of the
        object will be returned).

        If not, ``x`` is assumed to be an `ndarray`. The transformation will be
        non-destructive, returning the transformed version.

        Any ``kwargs`` will be passed to the specific transform :meth:`_apply`
        method.

        Parameters
        ----------
        x : :map:`Transformable` or ``(n_points, n_dims)`` `ndarray`
            The array or object to be transformed.
        batch_size : `int`, optional
            If not ``None``, this determines how many items from the numpy
            array will be passed through the transform at a time. This is
            useful for operations that require large intermediate matrices
            to be computed.
        kwargs : `dict`
            Passed through to :meth:`_apply`.

        Returns
        -------
        transformed : ``type(x)``
            The transformed object or array
        """

        def transform(x_):
            """
            Local closure which calls the :meth:`_apply` method with the
            `kwargs` attached.
            """
            return self._apply_batched(x_, batch_size, **kwargs)

        try:
            return x._transform(transform)
        except AttributeError:
            return self._apply_batched(x, batch_size, **kwargs)

    def _apply_batched(self, x, batch_size, **kwargs):
        if batch_size is None:
            return self._apply(x, **kwargs)
        else:
            outputs = []
            n_points = x.shape[0]
            for lo_ind in range(0, n_points, batch_size):
                hi_ind = lo_ind + batch_size
                outputs.append(self._apply(x[lo_ind:hi_ind], **kwargs))
            return np.vstack(outputs)

    def compose_before(self, transform):
        r"""
        Returns a :map:`TransformChain` that represents **this** transform
        composed **before** the given transform::

            c = a.compose_before(b)
            c.apply(p) == b.apply(a.apply(p))

        ``a`` and ``b`` are left unchanged.

        Parameters
        ----------
        transform : :map:`Transform`
            Transform to be applied **after** self

        Returns
        -------
        transform : :map:`TransformChain`
            The resulting transform chain.
        """
        return TransformChain([self, transform])

    def compose_after(self, transform):
        r"""
        Returns a :map:`TransformChain` that represents **this** transform
        composed **after** the given transform::

            c = a.compose_after(b)
            c.apply(p) == a.apply(b.apply(p))

        ``a`` and ``b`` are left unchanged.

        This corresponds to the usual mathematical formalism for the compose
        operator, `o`.

        Parameters
        ----------
        transform : :map:`Transform`
            Transform to be applied **before** self

        Returns
        -------
        transform : :map:`TransformChain`
            The resulting transform chain.
        """
        return TransformChain([transform, self])


class Transformable(Copyable):
    r"""
    Interface for objects that know how to be transformed by the
    :map:`Transform` interface.

    When ``Transform.apply_inplace`` is called on an object, the
    :meth:`_transform_inplace` method is called, passing in the transforms'
    :meth:`_apply` function.

    This allows for the object to define how it should transform itself.
    """

    def _transform_inplace(self, transform):
        r"""
        Apply the given transform function to ``self`` inplace.

        Parameters
        ----------
        transform : `function`
            Function that applies a transformation to the transformable object.

        Returns
        -------
        transformed : ``type(self)``
            The transformed object, having been transformed in place.
        """
        raise NotImplementedError()

    def _transform(self, transform):
        r"""
        Apply the :map:`Transform` given in a non destructive manner -
        returning the transformed object and leaving this object as it was.

        Parameters
        ----------
        transform : `function`
            Function that applies a transformation to the transformable object.

        Returns
        -------
        transformed : ``type(self)``
            A copy of the object, transformed.
        """
        copy_of_self = self.copy()
        # transform the copy destructively
        copy_of_self._transform_inplace(transform)
        return copy_of_self


from .alignment import Alignment
from .composable import TransformChain, ComposableTransform, VComposable
from .invertible import Invertible, VInvertible



from collections import OrderedDict
from mock import patch, MagicMock
import numpy as np
from pytest import raises
from scipy.sparse import csr_matrix

from menpo.image import Image
from menpo.shape import (TriMesh, TexturedTriMesh, ColouredTriMesh, PointCloud,
                         LabelledPointUndirectedGraph)
from menpo.visualize import Menpo3dMissingError
from menpo.testing import surrogate


fake_triangle = np.array([[0.0, 0.0, 0.0],
                          [1.0, 0.0, 0.0],
                          [0.0, 1.0, 0.0]])
fake_trilist = np.array([[0, 1, 2]], dtype=np.uint32)
fake_texture = Image.init_blank([10, 10])
fake_tcoords = np.array([[0, 0], [0.5, 0.5], [1.0, 1.0]])
menpo3d_visualize_mock = MagicMock()
menpo3d_visualize_mock.side_effect = ImportError


@surrogate('menpo3d.visualize.TriMeshViewer3d')
@patch('menpo3d.visualize.TriMeshViewer3d', menpo3d_visualize_mock)
def test_trimesh_viewer():
    with raises(Menpo3dMissingError):
        TriMesh(fake_triangle, trilist=fake_trilist, copy=False).view()


@surrogate('menpo3d.visualize.TexturedTriMeshViewer3d')
@patch('menpo3d.visualize.TexturedTriMeshViewer3d', menpo3d_visualize_mock)
def test_textured_trimesh_viewer():
    with raises(Menpo3dMissingError):
        TexturedTriMesh(fake_triangle, fake_tcoords, fake_texture,
                        trilist=fake_trilist, copy=False).view()


@surrogate('menpo3d.visualize.ColouredTriMeshViewer3d')
@patch('menpo3d.visualize.ColouredTriMeshViewer3d', menpo3d_visualize_mock)
def test_coloured_trimesh_viewer():
    with raises(Menpo3dMissingError):
        ColouredTriMesh(fake_triangle, colours=fake_tcoords,
                        trilist=fake_trilist, copy=False).view()


@surrogate('menpo3d.visualize.PointCloudViewer3d')
@patch('menpo3d.visualize.PointCloudViewer3d', menpo3d_visualize_mock)
def test_pointcloud3d_viewer():
    with raises(Menpo3dMissingError):
        PointCloud(fake_triangle, copy=False).view()


@surrogate('menpo3d.visualize.LandmarkViewer3d')
@patch('menpo3d.visualize.LandmarkViewer3d', menpo3d_visualize_mock)
def test_landmark3d_viewer():
    adj_matrix = csr_matrix((3, 3))
    labels_map = OrderedDict([('all', np.ones(3, dtype=np.bool))])
    with raises(Menpo3dMissingError):
        LabelledPointUndirectedGraph(fake_triangle,
                                     adj_matrix,
                                     labels_map,
                                     copy=False).view()



import numpy as np

from ..adjacency import mask_adjacency_array, reindex_adjacency_array
from .base import TriMesh


class ColouredTriMesh(TriMesh):
    r"""
    Combines a :map:`TriMesh` with a colour per vertex.

    Parameters
    ----------
    points : ``(n_points, n_dims)`` `ndarray`
        The array representing the points.
    trilist : ``(M, 3)`` `ndarray` or ``None``, optional
        The triangle list. If `None`, a Delaunay triangulation of
        the points will be used instead.
    colours : ``(N, 3)`` `ndarray`, optional
        The floating point RGB colour per vertex. If not given, grey will be
        assigned to each vertex.
    copy: `bool`, optional
        If ``False``, the points, trilist and colours will not be copied on
        assignment.
        In general this should only be used if you know what you are doing.

    Raises
    ------
    ValueError
        If the number of colour values does not match the number of vertices.
    """

    def __init__(self, points, trilist=None, colours=None, copy=True):
        TriMesh.__init__(self, points, trilist=trilist, copy=copy)
        # Handle the settings of colours, either be provided a default grey
        # set of colours, or copy the given array if necessary
        if colours is None:
            # default to grey
            colours_handle = np.ones_like(points, dtype=np.float) * 0.5
        elif not copy:
            colours_handle = colours
        else:
            colours_handle = colours.copy()

        if points.shape[0] != colours_handle.shape[0]:
            raise ValueError('Must provide a colour per-vertex.')
        self.colours = colours_handle

    @classmethod
    def init_2d_grid(cls, shape, spacing=None, colours=None):
        r"""
        Create a ColouredTriMesh that exists on a regular 2D grid. The first
        dimension is the number of rows in the grid and the second dimension
        of the shape is the number of columns. ``spacing`` optionally allows
        the definition of the distance between points (uniform over points).
        The spacing may be different for rows and columns.

        The triangulation will be right-handed and the diagonal will go from
        the top left to the bottom right of a square on the grid.

        Parameters
        ----------
        shape : `tuple` of 2 `int`
            The size of the grid to create, this defines the number of points
            across each dimension in the grid. The first element is the number
            of rows and the second is the number of columns.
        spacing : `int` or `tuple` of 2 `int`, optional
            The spacing between points. If a single `int` is provided, this
            is applied uniformly across each dimension. If a `tuple` is
            provided, the spacing is applied non-uniformly as defined e.g.
            ``(2, 3)`` gives a spacing of 2 for the rows and 3 for the
            columns.
        colours : ``(N, 3)`` `ndarray`, optional
            The floating point RGB colour per vertex. If not given, grey will be
            assigned to each vertex.

        Returns
        -------
        trimesh : :map:`TriMesh`
            A TriMesh arranged in a grid.
        """
        pc = TriMesh.init_2d_grid(shape, spacing=spacing)
        points = pc.points
        trilist = pc.trilist
        # Ensure that the colours are copied
        if colours is not None:
            colours = colours.copy()
        return ColouredTriMesh(points, trilist=trilist, colours=colours,
                               copy=False)

    @classmethod
    def init_from_depth_image(cls, depth_image, colours=None):
        r"""
        Return a 3D textured triangular mesh from the given depth image. The
        depth image is assumed to represent height/depth values and the XY
        coordinates are assumed to unit spaced and represent image coordinates.
        This is particularly useful for visualising depth values that have been
        recovered from images.

        The optionally passed texture will be textured mapped onto the planar
        surface using the correct texture coordinates for an image of the
        same shape as ``depth_image``.

        Parameters
        ----------
        depth_image : :map:`Image` or subclass
            A single channel image that contains depth values - as commonly
            returned by RGBD cameras, for example.
        colours : ``(N, 3)`` `ndarray`, optional
            The floating point RGB colour per vertex. If not given, grey will be
            assigned to each vertex.

        Returns
        -------
        depth_cloud : ``type(cls)``
            A new 3D TriMesh with unit XY coordinates and the given depth
            values as Z coordinates. The trilist is constructed as in
            :meth:`init_2d_grid`.
        """
        from menpo.image import MaskedImage

        new_tmesh = cls.init_2d_grid(depth_image.shape, colours=colours)
        if isinstance(depth_image, MaskedImage):
            new_tmesh = new_tmesh.from_mask(depth_image.mask.as_vector())
        return cls(np.hstack([new_tmesh.points,
                              depth_image.as_vector(keep_channels=True).T]),
                   colours=new_tmesh.colours,
                   trilist=new_tmesh.trilist,
                   copy=False)

    @property
    def n_channels(self):
        r"""
        The number of channels of colour used (e.g. 3 for RGB).

        :type: `int`
        """
        return self.colours.shape[1]

    def from_mask(self, mask):
        """
        A 1D boolean array with the same number of elements as the number of
        points in the ColouredTriMesh. This is then broadcast across the
        dimensions of the mesh and returns a new mesh containing only those
        points that were ``True`` in the mask.

        Parameters
        ----------
        mask : ``(n_points,)`` `ndarray`
            1D array of booleans

        Returns
        -------
        mesh : :map:`ColouredTriMesh`
            A new mesh that has been masked.
        """
        if mask.shape[0] != self.n_points:
            raise ValueError('Mask must be a 1D boolean array of the same '
                             'number of entries as points in this '
                             'ColouredTriMesh.')

        ctm = self.copy()
        if np.all(mask):  # Fast path for all true
            return ctm
        else:
            # Recalculate the mask to remove isolated vertices
            isolated_mask = self._isolated_mask(mask)
            # Recreate the adjacency array with the updated mask
            masked_adj = mask_adjacency_array(isolated_mask, self.trilist)
            ctm.trilist = reindex_adjacency_array(masked_adj)
            ctm.points = ctm.points[isolated_mask, :]
            ctm.colours = ctm.colours[isolated_mask, :]
            return ctm

    def clip_texture(self, range=(0., 1.)):
        """
        Method that returns a copy of the object with the coloured values
        clipped in range ``(0, 1)``.

        Parameters
        ----------
        range : ``(float, float)``, optional
            The clipping range.

        Returns
        -------
        self : :map:`ColouredTriMesh`
            A copy of self with its texture clipped.
        """
        instance = self.copy()
        instance.colours = np.clip(self.colours, *range)
        return instance

    def rescale_texture(self, minimum, maximum, per_channel=True):
        r"""
        A copy of this mesh with colours linearly rescaled to fit a range.

        Parameters
        ----------
        minimum: `float`
            The minimal value of the rescaled colours
        maximum: `float`
            The maximal value of the rescaled colours
        per_channel: `boolean`, optional
            If ``True``, each channel will be rescaled independently. If
            ``False``, the scaling will be over all channels.

        Returns
        -------
        coloured_mesh : ``type(self)``
            A copy of this mesh with colours linearly rescaled to fit in the
            range provided.
        """
        instance = self.copy()
        colours = instance.colours
        if per_channel:
            min_, max_ = colours.min(axis=0), colours.max(axis=0)
        else:
            min_, max_ = colours.min(), colours.max()
        sf = (maximum - minimum) / (max_ - min_)
        instance.colours = ((colours - min_) * sf) + minimum
        return instance

    def _view_3d(self, figure_id=None, new_figure=True, render_texture=True,
                 mesh_type='surface', ambient_light=0.0, specular_light=0.0,
                 colour='r', line_width=2, normals=None, normals_colour='k',
                 normals_line_width=2, normals_marker_style='2darrow',
                 normals_marker_resolution=8, normals_marker_size=None,
                 step=None, alpha=1.0):
        r"""
        Visualize the Coloured TriMesh in 3D.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        render_texture : `bool`, optional
            If ``True``, then the texture is rendered. If ``False``, then only
            the TriMesh is rendered with the specified `colour`.
        mesh_type : ``{'surface', 'wireframe'}``, optional
            The representation type to be used for the mesh.
        ambient_light : `float`, optional
            The ambient light intensity. It must be in range ``[0., 1.]``.
        specular_light : `float`, optional
            The specular light intensity. It must be in range ``[0., 1.]``.
        colour : See Below, optional
            The colour of the mesh if `render_texture` is ``False``.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        line_width : `float`, optional
            The width of the lines, if there are any.
        normals : ``(n_points, 3)`` `ndarray` or ``None``, optional
            If ``None``, then the normals will not be rendered. If `ndarray`,
            then the provided normals will be rendered as well. Note that a
            normal must be provided for each point in the TriMesh.
        normals_colour : See Below, optional
            The colour of the normals.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        normals_line_width : `float`, optional
            The width of the lines of the normals. It only applies if `normals`
            is not ``None``.
        normals_marker_style : `str`, optional
            The style of the markers of the normals. It only applies if `normals`
            is not ``None``.
            Example options ::

                {2darrow, 2dcircle, 2dcross, 2ddash, 2ddiamond, 2dhooked_arrow,
                 2dsquare, 2dthick_arrow, 2dthick_cross, 2dtriangle, 2dvertex,
                 arrow, axes, cone, cube, cylinder, point, sphere}

        normals_marker_resolution : `int`, optional
            The resolution of the markers of the normals. For spheres, for
            instance, this is the number of divisions along theta and phi. It
            only applies if `normals` is not ``None``.
        normals_marker_size : `float` or ``None``, optional
            The size of the markers. This size can be seen as a scale factor
            applied to the size markers, which is by default calculated from
            the inter-marker spacing. If ``None``, then an optimal marker size
            value will be set automatically. It only applies if `normals` is not
            ``None``.
        step : `int` or ``None``, optional
            If `int`, then one every `step` normals will be rendered.
            If ``None``, then all vertexes will be rendered. It only applies if
            `normals` is not ``None``.
        alpha : `float`, optional
            Defines the transparency (opacity) of the object.

        Returns
        -------
        renderer : `menpo3d.visualize.ColouredTriMeshViewer3D`
            The Menpo3D rendering object.
        """
        if render_texture:
            try:
                from menpo3d.visualize import ColouredTriMeshViewer3d
                renderer = ColouredTriMeshViewer3d(figure_id, new_figure,
                                                   self.points, self.trilist,
                                                   self.colours)
                renderer.render(
                    mesh_type=mesh_type, ambient_light=ambient_light,
                    specular_light=specular_light, normals=normals,
                    normals_colour=normals_colour,
                    normals_line_width=normals_line_width,
                    normals_marker_style=normals_marker_style,
                    normals_marker_resolution=normals_marker_resolution,
                    normals_marker_size=normals_marker_size, step=step,
                    alpha=alpha)
                return renderer
            except ImportError as e:
                from menpo.visualize import Menpo3dMissingError
                raise Menpo3dMissingError(e)
        else:
            try:
                from menpo3d.visualize import TriMeshViewer3d
                renderer = TriMeshViewer3d(figure_id, new_figure, self.points,
                                           self.trilist)
                renderer.render(
                    mesh_type=mesh_type, line_width=line_width, colour=colour,
                    normals=normals, normals_colour=normals_colour,
                    normals_line_width=normals_line_width,
                    normals_marker_style=normals_marker_style,
                    normals_marker_resolution=normals_marker_resolution,
                    normals_marker_size=normals_marker_size, step=step,
                    alpha=alpha)
                return renderer
            except ImportError as e:
                from menpo.visualize import Menpo3dMissingError
                raise Menpo3dMissingError(e)

    def _view_2d(self, figure_id=None, new_figure=False, image_view=True,
                 render_lines=True, line_colour='r', line_style='-',
                 line_width=1., render_markers=True, marker_style='o',
                 marker_size=5, marker_face_colour='k', marker_edge_colour='k',
                 marker_edge_width=1., render_numbering=False,
                 numbers_horizontal_align='center',
                 numbers_vertical_align='bottom',
                 numbers_font_name='sans-serif', numbers_font_size=10,
                 numbers_font_style='normal', numbers_font_weight='normal',
                 numbers_font_colour='k', render_axes=True,
                 axes_font_name='sans-serif', axes_font_size=10,
                 axes_font_style='normal', axes_font_weight='normal',
                 axes_x_limits=None, axes_y_limits=None, axes_x_ticks=None,
                 axes_y_ticks=None, figure_size=(7, 7), label=None, **kwargs):
        r"""
        Visualization of the TriMesh in 2D. Currently, explicit coloured TriMesh
        viewing is not supported, and therefore viewing falls back to uncoloured
        2D TriMesh viewing.

        Returns
        -------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        image_view : `bool`, optional
            If ``True`` the ColouredTriMesh will be viewed as if it is in the
            image coordinate system.
        render_lines : `bool`, optional
            If ``True``, the edges will be rendered.
        line_colour : See Below, optional
            The colour of the lines.
            Example options::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        line_style : ``{-, --, -., :}``, optional
            The style of the lines.
        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : See Below, optional
            The style of the markers. Example options ::

                {., ,, o, v, ^, <, >, +, x, D, d, s, p, *, h, H, 1, 2, 3, 4, 8}

        marker_size : `int`, optional
            The size of the markers in points.
        marker_face_colour : See Below, optional
            The face (filling) colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_colour : See Below, optional
            The edge colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_numbering : `bool`, optional
            If ``True``, the landmarks will be numbered.
        numbers_horizontal_align : ``{center, right, left}``, optional
            The horizontal alignment of the numbers' texts.
        numbers_vertical_align : ``{center, top, bottom, baseline}``, optional
            The vertical alignment of the numbers' texts.
        numbers_font_name : See Below, optional
            The font of the numbers. Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        numbers_font_size : `int`, optional
            The font size of the numbers.
        numbers_font_style : ``{normal, italic, oblique}``, optional
            The font style of the numbers.
        numbers_font_weight : See Below, optional
            The font weight of the numbers.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold, demibold, demi, bold, heavy, extra bold, black}

        numbers_font_colour : See Below, optional
            The font colour of the numbers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : See Below, optional
            The font of the axes.
            Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : {``normal``, ``italic``, ``oblique``}, optional
            The font style of the axes.
        axes_font_weight : See Below, optional
            The font weight of the axes.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold, demibold, demi, bold, heavy, extra bold, black}

        axes_x_limits : `float` or (`float`, `float`) or ``None``, optional
            The limits of the x axis. If `float`, then it sets padding on the
            right and left of the TriMesh as a percentage of the TriMesh's
            width. If `tuple` or `list`, then it defines the axis limits. If
            ``None``, then the limits are set automatically.
        axes_y_limits : (`float`, `float`) `tuple` or ``None``, optional
            The limits of the y axis. If `float`, then it sets padding on the
            top and bottom of the TriMesh as a percentage of the TriMesh's
            height. If `tuple` or `list`, then it defines the axis limits. If
            ``None``, then the limits are set automatically.
        axes_x_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the x axis.
        axes_y_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the y axis.
        figure_size : (`float`, `float`) `tuple` or ``None``, optional
            The size of the figure in inches.
        label : `str`, optional
            The name entry in case of a legend.

        Returns
        -------
        viewer : :map:`PointGraphViewer2d`
            The viewer object.

        Raises
        ------
        warning
            2D Viewing of Coloured TriMeshes is not supported, automatically
            falls back to 2D :map:`TriMesh` viewing.
        """
        import warnings
        warnings.warn(Warning('2D Viewing of Coloured TriMeshes is not '
                              'supported, falling back to TriMesh viewing.'))
        return TriMesh._view_2d(
            self, figure_id=figure_id, new_figure=new_figure,
            image_view=image_view, render_lines=render_lines,
            line_colour=line_colour, line_style=line_style,
            line_width=line_width, render_markers=render_markers,
            marker_style=marker_style, marker_size=marker_size,
            marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width,
            render_numbering=render_numbering,
            numbers_horizontal_align=numbers_horizontal_align,
            numbers_vertical_align=numbers_vertical_align,
            numbers_font_name=numbers_font_name,
            numbers_font_size=numbers_font_size,
            numbers_font_style=numbers_font_style,
            numbers_font_weight=numbers_font_weight,
            numbers_font_colour=numbers_font_colour, render_axes=render_axes,
            axes_font_name=axes_font_name, axes_font_size=axes_font_size,
            axes_font_style=axes_font_style, axes_font_weight=axes_font_weight,
            axes_x_limits=axes_x_limits, axes_y_limits=axes_y_limits,
            axes_x_ticks=axes_x_ticks, axes_y_ticks=axes_y_ticks,
            figure_size=figure_size, label=label)



import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
from pytest import raises

from menpo.transform import UniformScale, NonUniformScale, Scale


def test_nonuniformscale_from_list():
    u_a = NonUniformScale([3, 2, 3])
    u_b = NonUniformScale(np.array([3, 2, 3]))
    assert (np.all(u_a.h_matrix == u_b.h_matrix))


def test_uniformscale2d_n_parameters():
    scale = 2
    t = UniformScale(scale, 2)
    assert (t.n_parameters == 1)


def test_uniformscale3d_n_parameters():
    scale = 2
    t = UniformScale(scale, 3)
    assert (t.n_parameters == 1)


def test_nonuniformscale_2d_n_parameters():
    scale = np.array([1, 2])
    t = NonUniformScale(scale)
    assert (t.n_parameters == 2)


def test_uniformscale2d_update_from_vector():
    # make a uniform scale of 1, 2 dimensional
    uniform_scale = UniformScale(1, 2)
    new_scale = 2
    homo = np.array([[new_scale, 0, 0],
                     [0, new_scale, 0],
                     [0, 0, 1]])

    uniform_scale._from_vector_inplace(new_scale)
    assert_almost_equal(uniform_scale.h_matrix, homo)


def test_uniformscale2d_as_vector():
    scale = 2
    vec = UniformScale(scale, 2).as_vector()
    assert_allclose(vec, scale)


def test_nonuniformscale2d_from_vector():
    scale = np.array([1, 2])
    homo = np.array([[scale[0], 0, 0],
                     [0, scale[1], 0],
                     [0, 0, 1]])

    tr = NonUniformScale.init_identity(2).from_vector(scale)

    assert_almost_equal(tr.h_matrix, homo)


def test_nonuniformscale2d_update_from_vector():
    scale = np.array([3, 4])
    homo = np.array([[scale[0], 0, 0],
                     [0, scale[1], 0],
                     [0, 0, 1]])
    tr = NonUniformScale(np.array([1, 2]))
    tr._from_vector_inplace(scale)
    assert_almost_equal(tr.h_matrix, homo)


def test_nonuniformscale2d_as_vector():
    scale = np.array([1, 2])
    vec = NonUniformScale(scale).as_vector()
    assert_allclose(vec, scale)


def test_scale_2d_pseudoinverse():
    scale1 = 0.5
    scale2 = 4.0
    homo = np.array([[scale1,      0, 0],
                     [     0, scale2, 0],
                     [     0,      0, 1]])

    tr = NonUniformScale([1/scale1, 1/scale2])
    assert_almost_equal(tr.pseudoinverse().h_matrix, homo)


def test_uniformscale3d_from_vector():
    scale = 2
    homo = np.array([[scale, 0, 0, 0],
                     [0, scale, 0, 0],
                     [0, 0, scale, 0],
                     [0, 0, 0, 1]])

    uniform_scale = UniformScale(1, 3)
    tr = uniform_scale.from_vector(scale)
    assert_almost_equal(tr.h_matrix, homo)


def test_uniformscale3d_as_vector():
    scale = 2
    vec = UniformScale(scale, 3).as_vector()
    assert_allclose(vec, scale)


def test_uniformscale_build_2d():
    scale = 2
    homo = np.array([[scale, 0, 0],
                     [0, scale, 0],
                     [0, 0, 1]])

    tr = UniformScale(scale, 2)
    assert_almost_equal(tr.h_matrix, homo)


def test_uniformscale_build_3d():
    scale = 2
    homo = np.array([[scale, 0, 0, 0],
                     [0, scale, 0, 0],
                     [0, 0, scale, 0],
                     [0, 0, 0, 1]])

    tr = UniformScale(scale, 3)

    assert (isinstance(tr, UniformScale))
    assert_almost_equal(tr.h_matrix, homo)


def test_uniformscale_build_4d_raise_dimensionalityerror():
    with raises(ValueError):
        UniformScale(1, 4)


def test_uniformscale_2d_pseudoinverse():
    scale = 0.5
    homo = np.array([[scale, 0, 0],
                     [0, scale, 0],
                     [0, 0, 1]])

    tr = UniformScale(2, 2)
    assert_almost_equal(tr.pseudoinverse().h_matrix, homo)


def test_scale_build_2d_uniform_pass_dim():
    scale = 2
    ndim = 2
    tr = Scale(scale, ndim)

    assert (isinstance(tr, UniformScale))


def test_scale_build_3d_uniform_pass_dim():
    scale = 2
    ndim = 3
    tr = Scale(scale, ndim)

    assert (isinstance(tr, UniformScale))


def test_scale_build_2d_nonuniform():
    scale = np.array([1, 2])
    tr = Scale(scale)

    assert (isinstance(tr, NonUniformScale))


def test_scale_build_2d_uniform_from_vec():
    scale = np.array([2, 2])
    tr = Scale(scale)

    assert (isinstance(tr, UniformScale))


def test_scale_zero_scale_raise_valuerror():
    with raises(ValueError):
        Scale(np.array([1, 0]))


def test_uniformscale_identity_2d():
    assert_allclose(UniformScale.init_identity(2).h_matrix, np.eye(3))


def test_uniformscale_identity_3d():
    assert_allclose(UniformScale.init_identity(3).h_matrix, np.eye(4))


def test_nonuniformscale_identity_2d():
    assert_allclose(NonUniformScale.init_identity(2).h_matrix, np.eye(3))


def test_nonuniformscale_identity_3d():
    assert_allclose(NonUniformScale.init_identity(3).h_matrix, np.eye(4))



import numpy as np
from numpy.testing import assert_allclose
from menpo.shape import PointCloud, mean_pointcloud, TriMesh


def test_mean_pointcloud():
    points = np.array([[1, 2, 3],
                       [1, 1, 1]])
    pcs = [PointCloud(points), PointCloud(points + 2)]
    mean_pc = mean_pointcloud(pcs)
    assert isinstance(mean_pc, PointCloud)
    assert_allclose(mean_pc.points, points + 1)


def test_mean_pointcloud_type():
    points = np.array([[1, 2, 3],
                       [1, 1, 1]])
    trilist = np.array([0, 1, 2])
    pcs = [TriMesh(points, trilist), TriMesh(points + 2, trilist)]
    mean_pc = mean_pointcloud(pcs)
    assert isinstance(mean_pc, TriMesh)
    assert_allclose(mean_pc.points, points + 1)



import numpy as np

from ..homogeneous import AlignmentSimilarity
from .base import MultipleAlignment


avoid_circular = None      # to avoid circular imports
mean_pointcloud = None     # to avoid circular imports
PointCloud = None          # to avoid circular imports
scale_about_centre = None  # to avoid circular imports


class GeneralizedProcrustesAnalysis(MultipleAlignment):
    r"""
    Class for aligning multiple source shapes between them.

    After construction, the :map:`AlignmentSimilarity` transforms used to map
    each `source` optimally to the `target` can be found at `transforms`.

    Parameters
    ----------
    sources : `list` of :map:`PointCloud`
        List of pointclouds to be aligned.
    target : :map:`PointCloud`, optional
        The target :map:`PointCloud` to align each source to.
        If ``None``, then the mean of the sources is used.
    allow_mirror : `bool`, optional
        If ``True``, the Kabsch algorithm check is not performed, and mirroring
        of the Rotation matrix is permitted.

    Raises
    ------
    ValueError
        Need at least two sources to align
    """
    def __init__(self, sources, target=None, allow_mirror=False):
        super(GeneralizedProcrustesAnalysis, self).__init__(sources,
                                                            target=target)
        initial_target = self.target
        self.transforms = [AlignmentSimilarity(source, self.target,
                                               allow_mirror=allow_mirror)
                           for source in self.sources]
        self.initial_target_scale = self.target.norm()
        self.n_iterations = 1
        self.max_iterations = 100
        self.converged = self._recursive_procrustes()
        if target is not None:
            self.target = initial_target

    def _recursive_procrustes(self):
        r"""
        Recursively calculates a procrustes alignment.
        """
        global mean_pointcloud, PointCloud, scale_about_centre, avoid_circular
        if avoid_circular is None:
            from menpo.shape import mean_pointcloud, PointCloud
            from ..compositions import scale_about_centre
            avoid_circular = True

        if self.n_iterations > self.max_iterations:
            return False
        new_tgt = mean_pointcloud([PointCloud(t.aligned_source().points,
                                              copy=False)
                                   for t in self.transforms])
        # rescale the new_target to be the same size as the original about
        # it's centre
        rescale = scale_about_centre(new_tgt,
                                     self.initial_target_scale / new_tgt.norm())
        rescale._apply_inplace(new_tgt)
        # check to see if we have converged yet
        delta_target = np.linalg.norm(self.target.points - new_tgt.points)
        if delta_target < 1e-6:
            return True
        else:
            self.n_iterations += 1
            for t in self.transforms:
                t.set_target(new_tgt)
            self.target = new_tgt
            return self._recursive_procrustes()

    def mean_aligned_shape(self):
        r"""
        Returns the mean of the aligned shapes.

        :type: :map:`PointCloud`
        """
        from menpo.shape import PointCloud
        return PointCloud(np.mean([t.target.points for t in self.transforms],
                                  axis=0))

    def mean_alignment_error(self):
        r"""
        Returns the average error of the recursive procrustes alignment.

        :type: `float`
        """
        return sum([t.alignment_error() for t in
                    self.transforms])/self.n_sources

    def __str__(self):
        if self.converged:
            return ('Converged after %d iterations with av. error %f'
                    % (self.n_iterations, self.mean_alignment_error()))
        else:
            return ('Failed to converge after %d iterations with av. error '
                    '%f' % (self.n_iterations, self.mean_alignment_error()))



import numpy as np
from numpy.testing import assert_array_equal

import menpo.io as mio
from menpo.image.base import (Image, _convert_patches_list_to_single_array,
                              _create_patches_image)
from menpo.shape import PointCloud


#######################
# EXTRACT PATCHES TESTS
#######################
def test_double_type():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert(patches[0].pixels.dtype == np.float64)


def test_float_type():
    image = mio.import_builtin_asset('breakingbad.jpg')
    image.pixels = image.pixels.astype(np.float32)
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert(patches[0].pixels.dtype == np.float32)


def test_uint8_type():
    image = mio.import_builtin_asset('breakingbad.jpg', normalize=False)
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert(patches[0].pixels.dtype == np.uint8)


def test_uint16_type():
    image = Image.init_blank([100, 100], dtype=np.uint16)
    patch_shape = (16, 16)
    landmarks = PointCloud(np.array([[50, 50.]]))
    patches = image.extract_patches(landmarks,
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert(patches[0].pixels.dtype == np.uint16)


def test_int_pointcloud():
    image = Image.init_blank([100, 100])
    patch_shape = (16, 16)
    landmarks = PointCloud(np.array([[50, 50]]))
    patches = image.extract_patches(landmarks,
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert(patches[0].pixels.dtype == np.float)


def test_uint8_type_single_array():
    image = mio.import_builtin_asset('breakingbad.jpg', normalize=False)
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=True)
    assert(patches.dtype == np.uint8)


def test_squared_even_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert len(patches) == 68


def test_squared_odd_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (15, 15)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert len(patches) == 68


def test_nonsquared_even_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 18)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert len(patches) == 68


def test_nonsquared_odd_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (15, 17)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert len(patches) == 68


def test_nonsquared_even_odd_patches():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (15, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    patch_shape=patch_shape,
                                    as_single_array=False)
    assert len(patches) == 68


def test_squared_even_patches_landmarks():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 16)
    patches = image.extract_patches_around_landmarks('PTS',
                                                     patch_shape=patch_shape,
                                                     as_single_array=False)
    assert len(patches) == 68


def test_squared_even_patches_single_array():
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_shape = (16, 16)
    patches = image.extract_patches(image.landmarks['PTS'],
                                    as_single_array=True,
                                    patch_shape=patch_shape)
    assert patches.shape, ((68, 1 == 3) + patch_shape)


def test_squared_even_patches_sample_offsets():
    image = mio.import_builtin_asset('breakingbad.jpg')
    sample_offsets = np.array([[0, 0], [1, 0]])
    patches = image.extract_patches(image.landmarks['PTS'],
                                    sample_offsets=sample_offsets,
                                    as_single_array=False)
    assert len(patches) == 136


#######################
# SET PATCHES TESTS
#######################
def test_single_ndarray_patch():
    patch_shape = (21, 7)
    n_channels = 4
    im = Image.init_blank(patch_shape, n_channels)
    patch = np.zeros((2, 2, n_channels) + patch_shape)
    patch[1, 0, ...] = np.ones((n_channels,) + patch_shape)
    patch[1, 1, ...] = 2 * np.ones((n_channels,) + patch_shape)
    patch_center = PointCloud(np.array([[10., 3.], [11., 3.]]))
    new_im = im.set_patches(patch, patch_center, offset=(0, 0), offset_index=1)
    res = np.zeros(patch_shape)
    res[1:-1, :] = 2
    assert_array_equal(new_im.pixels[2, ...], res)


def test_single_list_patch():
    patch_shape = (21, 7)
    n_channels = 4
    im = Image.init_blank(patch_shape, n_channels)
    patch = [Image(np.ones((n_channels,) + patch_shape)),
             Image(2 * np.ones((n_channels,) + patch_shape))]
    patch_center = PointCloud(np.array([[10., 3.], [11., 3.]]))
    new_im = im.set_patches(patch, patch_center, offset=(0, 0), offset_index=0)
    res = np.ones(patch_shape)
    res[1:-1, :] = 2
    assert_array_equal(new_im.pixels[2, ...], res)


def test_offset_argument():
    patch_shape = (5, 6)
    offsets = [(0., 0.), [0., 0.], np.array([[1., 1.]]), None]
    image = mio.import_builtin_asset('breakingbad.jpg')
    patch_center = PointCloud(np.array([[100., 101.], [50., 41.]]))
    patch = np.zeros((2, 1, image.n_channels) + patch_shape)
    patch[0, 0, ...] = np.ones((image.n_channels,) + patch_shape)
    patch[1, 0, ...] = 2 * np.ones((image.n_channels,) + patch_shape)
    for off in offsets:
        image = image.set_patches(patch, patch_center, offset=off)
        assert_array_equal(image.pixels[:, 98:103, 98:104], patch[0, 0, ...])
        assert_array_equal(image.pixels[:, 48:53, 38:44], patch[1, 0, ...])


def test_convert_patches_list_to_single_array():
    patch_shape = (7, 2)
    n_channels = 10
    n_centers = 2
    n_offsets = 2
    patches_list = [Image(1 * np.ones((n_channels,) + patch_shape)),
                    Image(2 * np.ones((n_channels,) + patch_shape)),
                    Image(3 * np.ones((n_channels,) + patch_shape)),
                    Image(4 * np.ones((n_channels,) + patch_shape))]
    patches_array = np.zeros((n_centers, n_offsets, n_channels) + patch_shape)
    patches_array[0, 0, ...] = patches_list[0].pixels
    patches_array[0, 1, ...] = patches_list[1].pixels
    patches_array[1, 0, ...] = patches_list[2].pixels
    patches_array[1, 1, ...] = patches_list[3].pixels
    assert_array_equal(
        _convert_patches_list_to_single_array(patches_list, n_centers),
        patches_array)


def test_set_patches_around_landmarks():
    patch_shape = (21, 12)
    image = mio.import_builtin_asset.takeo_ppm()
    patches1 = image.extract_patches_around_landmarks(
        patch_shape=patch_shape, as_single_array=True)
    new_image1 = Image.init_blank(image.shape, image.n_channels)
    new_image1.landmarks['PTS'] = image.landmarks['PTS']
    extracted1 = new_image1.set_patches_around_landmarks(patches1)

    patches2 = image.extract_patches_around_landmarks(
        patch_shape=patch_shape, as_single_array=False)
    new_image2 = Image.init_blank(image.shape, image.n_channels)
    new_image2.landmarks['PTS'] = image.landmarks['PTS']
    extracted2 = new_image2.set_patches_around_landmarks(patches2)
    assert_array_equal(extracted1.pixels, extracted2.pixels)


def test_create_patches_image():
    patch_shape = (7, 14)
    image = mio.import_builtin_asset.takeo_ppm()
    patches = image.extract_patches_around_landmarks(
        patch_shape=patch_shape, as_single_array=True)
    pc = image.landmarks['PTS']
    patches_image = _create_patches_image(patches, pc,
                                          patches_indices=list(range(17)))
    assert(patches_image.n_channels == patches.shape[2])
    assert(patches_image.landmarks.n_groups == 1)
    assert(patches_image.landmarks['patch_centers'].n_points == 17)



from itertools import islice
import numpy as np
from menpo.visualize import print_progress, bytes_str, print_dynamic


def dot_inplace_left(a, b, block_size=1000):
    r"""
    Inplace dot product for memory efficiency. It computes ``a * b = c``, where
    ``a`` will be replaced inplace with ``c``.

    Parameters
    ----------
    a : ``(n_big, k)`` `ndarray`
        First array to dot - assumed to be large. Will be damaged by this
        function call as it is used to store the output inplace.
    b : ``(k, n_small)`` `ndarray`, ``n_small <= k``
        The second array to dot - assumed to be small. ``n_small`` must be
        smaller than ``k`` so the result can be stored within the memory space
        of ``a``.
    block_size : `int`, optional
        The size of the block of ``a`` that will be dotted against ``b`` in
        each iteration. larger block sizes increase the time performance of the
        dot product at the cost of a higher memory overhead for the operation.

    Returns
    -------
    c : ``(n_big, n_small)`` `ndarray`
        The output of the operation. Exactly the same as a memory view onto
        ``a`` (``a[:, :n_small]``) as ``a`` is modified inplace to store the
        result.
    """
    (n_big, k_a), (k_b, n_small) = a.shape, b.shape
    if k_a != k_b:
        raise ValueError('Cannot dot {} * {}'.format(a.shape, b.shape))
    if n_small > k_a:
        raise ValueError('Cannot dot inplace left - '
                         'b.shape[1] ({}) > a.shape[1] '
                         '({})'.format(n_small, k_a))
    for i in range(0, n_big, block_size):
        j = i + block_size
        a[i:j, :n_small] = a[i:j].dot(b)
    return a[:, :n_small]


def dot_inplace_right(a, b, block_size=1000):
    r"""
    Inplace dot product for memory efficiency. It computes ``a * b = c`` where
    ``b`` will be replaced inplace with ``c``.

    Parameters
    ----------
    a : ``(n_small, k)`` `ndarray`, n_small <= k
        The first array to dot - assumed to be small. ``n_small`` must be
        smaller than ``k`` so the result can be stored within the memory space
        of ``b``.
    b : ``(k, n_big)`` `ndarray`
        Second array to dot - assumed to be large. Will be damaged by this
        function call as it is used to store the output inplace.
    block_size : `int`, optional
        The size of the block of ``b`` that ``a`` will be dotted against
        in each iteration. larger block sizes increase the time performance of
        the dot product at the cost of a higher memory overhead for the
        operation.

    Returns
    -------
    c : ``(n_small, n_big)`` `ndarray`
        The output of the operation. Exactly the same as a memory view onto
        ``b`` (``b[:n_small]``) as ``b`` is modified inplace to store the
        result.
    """
    (n_small, k_a), (k_b, n_big) = a.shape, b.shape
    if k_a != k_b:
        raise ValueError('Cannot dot {} * {}'.format(a.shape, b.shape))
    if n_small > k_b:
        raise ValueError('Cannot dot inplace right - '
                         'a.shape[1] ({}) > b.shape[0] '
                         '({})'.format(n_small, k_b))
    for i in range(0, n_big, block_size):
        j = i + block_size
        b[:n_small, i:j] = a.dot(b[:, i:j])
    return b[:n_small]


def as_matrix(vectorizables, length=None, return_template=False, verbose=False):
    r"""
    Create a matrix from a list/generator of :map:`Vectorizable` objects.
    All the objects in the list **must** be the same size when vectorized.

    Consider using a generator if the matrix you are creating is large and
    passing the length of the generator explicitly.

    Parameters
    ----------
    vectorizables : `list` or generator if :map:`Vectorizable` objects
        A list or generator of objects that supports the vectorizable interface
    length : `int`, optional
        Length of the vectorizable list. Useful if you are passing a generator
        with a known length.
    verbose : `bool`, optional
        If ``True``, will print the progress of building the matrix.
    return_template : `bool`, optional
        If ``True``, will return the first element of the list/generator, which
        was used as the template. Useful if you need to map back from the
        matrix to a list of vectorizable objects.

    Returns
    -------
    M : (length, n_features) `ndarray`
        Every row is an element of the list.
    template : :map:`Vectorizable`, optional
        If ``return_template == True``, will return the template used to
        build the matrix `M`.

    Raises
    ------
    ValueError
        ``vectorizables`` terminates in fewer than ``length`` iterations
    """
    # get the first element as the template and use it to configure the
    # data matrix
    if length is None:
        # samples is a list
        length = len(vectorizables)
        template = vectorizables[0]
        vectorizables = vectorizables[1:]
    else:
        # samples is an iterator
        template = next(vectorizables)
    n_features = template.n_parameters
    template_vector = template.as_vector()

    data = np.zeros((length, n_features), dtype=template_vector.dtype)
    if verbose:
        print('Allocated data matrix of size {} '
              '({} samples)'.format(bytes_str(data.nbytes), length))

    # now we can fill in the first element from the template
    data[0] = template_vector
    del template_vector

    # ensure we take at most the remaining length - 1 elements
    vectorizables = islice(vectorizables, length - 1)

    if verbose:
        vectorizables = print_progress(vectorizables, n_items=length, offset=1,
                                       prefix='Building data matrix',
                                       end_with_newline=False)

    # 1-based as we have the template vector set already
    i = 0
    for i, sample in enumerate(vectorizables, 1):
        data[i] = sample.as_vector()

    # we have exhausted the iterable, but did we get enough items?
    if i != length - 1:  # -1
        raise ValueError('Incomplete data matrix due to early iterator '
                         'termination (expected {} items, got {})'.format(
            length, i + 1))

    if return_template:
        return data, template
    else:
        return data


def from_matrix(matrix, template):
    r"""
    Create a generator from a matrix given a template :map:`Vectorizable`
    objects as a template. The ``from_vector`` method will be used to
    reconstruct each object.

    If you want a list, warp the returned value in ``list()``.

    Parameters
    ----------
    matrix : (n_items, n_features) `ndarray`
        A matrix whereby every *row* represents the data of a vectorizable
        object.
    template : :map:`Vectorizable`
        The template object to use to reconstruct each row of the matrix with.

    Returns
    -------
    vectorizables : generator of :map:`Vectorizable`
        Every row of the matrix becomes an element of the list.
    """
    return (template.from_vector(row) for row in matrix)



try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping
import fnmatch
from collections import OrderedDict

from menpo.base import Copyable
from menpo.transform.base import Transformable
from menpo.visualize.base import viewwrapper


class Landmarkable(Copyable):
    r"""
    Abstract interface for object that can have landmarks attached to them.
    Landmarkable objects have a public dictionary of landmarks which are
    managed by a :map:`LandmarkManager`. This means that
    different sets of landmarks can be attached to the same object.
    Landmarks can be N-dimensional and are expected to be some
    subclass of :map:`PointCloud` or :map:`LabelledPointUndirectedGraph`.
    """

    def __init__(self):
        self._landmarks = None

    def n_dims(self):
        """
        The total number of dimensions.

        :type: `int`
        """
        raise NotImplementedError()

    @property
    def landmarks(self):
        """
        The landmarks object.

        :type: :map:`LandmarkManager`
        """
        if self._landmarks is None:
            self._landmarks = LandmarkManager()
        return self._landmarks

    @property
    def has_landmarks(self):
        """
        Whether the object has landmarks.

        :type: `bool`
        """
        return self._landmarks is not None and self.landmarks.n_groups != 0

    @landmarks.setter
    def landmarks(self, value):
        """
        Landmarks setter.

        Parameters
        ----------
        value : :map:`LandmarkManager`
            The landmarks to set.
        """
        # firstly, make sure the dim is correct. Note that the dim can be None
        lm_n_dims = value.n_dims
        if lm_n_dims is not None and lm_n_dims != self.n_dims:
            raise ValueError(
                "Trying to set {}D landmarks on a "
                "{}D object".format(value.n_dims, self.n_dims))
        self._landmarks = value.copy()

    @property
    def n_landmark_groups(self):
        r"""
        The number of landmark groups on this object.

        :type: `int`
        """
        return self.landmarks.n_groups


class LandmarkManager(MutableMapping, Transformable):
    """Store for :map:`PointCloud` or ::map:`LabelledPointUndirectedGraph`
    instances associated with an object.

    Every :map:`Landmarkable` instance has an instance of this class available
    at the ``.landmarks`` property.  It is through this class that all access
    to landmarks attached to instances is handled. In general the
    :map:`LandmarkManager` provides a dictionary-like interface for storing
    landmarks. The LandmarkManager will contain instances of :map:`PointCloud`
    or :map:`LabelledPointUndirectedGraph` or subclasses thereof.
    :map:`LabelledPointUndirectedGraph` is unique in it's ability to
    include labels that refer to subsets of the underlying points that represent
    interesting semantic *labels*. These :map:`PointCloud` or
    :map:`LabelledPointUndirectedGraph` (or subclasses) are stored under
    string keys - these keys are refereed to as the **group name**. A special
    case is where there is a single unambiguous group attached to a
    :map:`LandmarkManager` - in this case ``None`` can be used as a key to
    access this sole group.

    Note that all groups stored on a :map:`Landmarkable` in it's attached
    :map:`LandmarkManager` are automatically transformed and copied with their
    parent object.
    """

    def __init__(self):
        super(LandmarkManager, self).__init__()
        self._landmark_groups = OrderedDict()

    @property
    def n_dims(self):
        """
        The total number of dimensions.

        :type: `int`
        """
        if self.n_groups != 0:
            # Python version independent way of getting the first value
            for v in self._landmark_groups.values():
                return v.n_dims
        else:
            return None

    def copy(self):
        r"""
        Generate an efficient copy of this :map:`LandmarkManager`.

        Returns
        -------
        ``type(self)``
            A copy of this object
        """
        # The dict will be shallow copied - rectify that here
        new = Copyable.copy(self)
        for k, v in new._landmark_groups.items():
            new._landmark_groups[k] = v.copy()
        return new

    def __iter__(self):
        """
        Iterate over the internal landmark group dictionary.
        """
        return iter(self._landmark_groups)

    def __setitem__(self, group, value):
        """
        Sets a new landmark group for the given label. This can be set using
        any :map`PointCloud` subclass. Existing landmark groups will be
        replaced.

        Parameters
        ----------
        group : `string`
            Label of new group.
        value : :map:`PointCloud` or subclass
            The new landmark group to set.

        Raises
        ------
        DimensionalityError
            If the landmarks and the shape are not of the same dimensionality.
        """
        from menpo.shape import PointCloud
        if group is None:
            raise ValueError('Cannot set using the key `None`. `None` has a '
                             'reserved meaning for landmark groups.')

        # firstly, make sure the dim is correct
        n_dims = self.n_dims
        if n_dims is not None and value.n_dims != n_dims:
            raise ValueError(
                "Trying to set {}D landmarks on a "
                "{}D LandmarkManager".format(value.n_dims, self.n_dims))
        if not isinstance(value, PointCloud):
            raise ValueError('Valid types are any subclass of PointCloud')

        # Copy the landmark group so that we now own it
        lmark_group = value.copy()
        self._landmark_groups[group] = lmark_group

    def __getitem__(self, group=None):
        """
        Returns the group for the provided label.

        Parameters
        ---------
        group : `string`, optional
            The label of the group. If None is provided, and if there is only
            one group, the unambiguous group will be returned.

        Returns
        -------
        lmark_group : :map:`PointCloud` or :map:`LabelledPointUndirectedGraph`
            The matching landmarks.
        """
        if group is None:
            if self.n_groups == 1:
                group = self.group_labels[0]
            else:
                raise ValueError("Cannot use None as a key as there are {} "
                                 "landmark groups".format(self.n_groups))
        return self._landmark_groups[group]

    def __delitem__(self, group):
        """
        Delete the group for the provided label.

        Parameters
        ---------
        group : `string`
            The label of the group.
        """
        del self._landmark_groups[group]

    def __len__(self):
        return len(self._landmark_groups)

    def __setstate__(self, state):
        # consistency with older versions imported.
        if not isinstance(state['_landmark_groups'], OrderedDict):
            state['_landmark_groups'] = OrderedDict(state['_landmark_groups'])

        self.__dict__ = state

    def _ipython_key_completions_(self):
        # Opt in to IPython tab completion - see
        # https://github.com/ipython/ipython/blob/5.2.2/docs/source/config/integrating.rst
        return list(self)

    @property
    def n_groups(self):
        """
        Total number of labels.

        :type: `int`
        """
        return len(self._landmark_groups)

    @property
    def has_landmarks(self):
        """
        Whether the object has landmarks or not

        :type: `int`
        """
        return self.n_groups != 0

    @property
    def group_labels(self):
        """
        All the labels for the landmark set sorted by insertion order.

        :type: `list` of `str`
        """
        # Convert to list so that we can index immediately, as keys()
        # is a generator in Python 3
        return list(self._landmark_groups.keys())

    def keys_matching(self, glob_pattern):
        r"""
        Yield only landmark group names (keys) matching a given glob.

        Parameters
        ----------
        glob_pattern : `str`
            A glob pattern e.g. 'frontal_face_*'

        Yields
        ------
        keys: group labels that match the glob pattern
        """
        for key in fnmatch.filter(self.keys(), glob_pattern):
            yield key

    def items_matching(self, glob_pattern):
        r"""
        Yield only items ``(group, PointCloud)`` where the key matches a
        given glob.

        Parameters
        ----------
        glob_pattern : `str`
            A glob pattern e.g. 'frontal_face_*'

        Yields
        ------
        item : ``(group, PointCloud)``
            Tuple of (str, PointCloud) where the group matches the glob.
        """
        for k, v in self.items():
            if fnmatch.fnmatch(k, glob_pattern):
                yield k, v

    def _transform_inplace(self, transform):
        for group in self._landmark_groups.values():
            group._transform_inplace(transform)
        return self

    @viewwrapper
    def view_widget(self):
        r"""
        Abstract method for viewing with an interactive widget. See the
        :map:`viewwrapper` documentation for an explanation of how the
        `view_widget` method works.
        """
        pass

    def _view_widget_2d(self, figure_size=(7, 7)):
        r"""
        Visualization of the landmark manager using an interactive widget.

        Parameters
        ----------
        figure_size : (`int`, `int`), optional
            The initial size of the rendered figure.
        """
        try:
            from menpowidgets import view_widget
            view_widget(self, figure_size=figure_size)
        except ImportError as e:
            from menpo.visualize.base import MenpowidgetsMissingError
            raise MenpowidgetsMissingError(e)

    def _view_widget_3d(self):
        r"""
        Visualization of the landmark manager using an interactive widget.
        """
        try:
            from menpowidgets import view_widget
            view_widget(self)
        except ImportError as e:
            from menpo.visualize.base import MenpowidgetsMissingError
            raise MenpowidgetsMissingError(e)

    def __str__(self):
        out_string = '{}: n_groups: {}'.format(type(self).__name__,
                                               self.n_groups)
        if self.has_landmarks:
            for label in self:
                out_string += '\n'
                out_string += '({}): {}'.format(label, self[label].__str__())

        return out_string


# TODO: Deprecate this - this handles importing old-style LandmarkGroup
class LandmarkGroup(object):

    def __new__(cls, *args, **kwargs):
        # This is a crazy hack that means when old style LandmarkGroup
        # objects are imported it is actually new-style
        # LabelledPointUndirectedGraph objects that are created.
        from menpo.shape import LabelledPointUndirectedGraph
        return LabelledPointUndirectedGraph.__new__(
            LabelledPointUndirectedGraph, *args, **kwargs)



from .linear import (LinearVectorModel, MeanLinearVectorModel, LinearModel,
                     MeanLinearModel)
from .pca import PCAModel, PCAVectorModel
from .gmrf import GMRFModel, GMRFVectorModel



import numpy as np
from numpy.testing import assert_allclose, raises
from menpo.shape import PointCloud
from menpo.transform import (Affine, AlignmentAffine,
                             Similarity, AlignmentSimilarity,
                             Rotation, AlignmentRotation,
                             Translation, AlignmentTranslation,
                             UniformScale, AlignmentUniformScale)

# TODO check composition works correctly on all alignment methods


# AFFINE

def test_align_2d_affine():
    linear_component = np.array([[1, -6],
                                 [-3, 2]])
    translation_component = np.array([7, -8])
    h_matrix = np.eye(3, 3)
    h_matrix[:-1, :-1] = linear_component
    h_matrix[:-1, -1] = translation_component
    affine = Affine(h_matrix)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = affine.apply(source)
    # estimate the transform from source and target
    estimate = AlignmentAffine(source, target)
    # check the estimates is correct
    assert_allclose(affine.h_matrix, estimate.h_matrix, atol=1e-3, rtol=1e-3)


def test_align_2d_affine_compose_target():
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = UniformScale(2.0, n_dims=2).apply(source)
    original_estimate = AlignmentAffine(source, target)
    new_estimate = original_estimate.copy()
    new_estimate.compose_after_from_vector_inplace(
        np.array([0, 0, 0, 0, 1, 1.]))
    estimate_target = new_estimate.target

    correct_target = original_estimate.compose_after(
        Translation([1, 1.])).apply(source)

    assert_allclose(estimate_target.points, correct_target.points)


def test_align_2d_affine_set_target():
    linear_component = np.array([[1, -6],
                                 [-3, 2]])
    translation_component = np.array([7, -8])
    h_matrix = np.eye(3, 3)
    h_matrix[:-1, :-1] = linear_component
    h_matrix[:-1, -1] = translation_component
    affine = Affine(h_matrix)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = affine.apply(source)
    # estimate the transform from source and source
    estimate = AlignmentAffine(source, source)
    # and set the target
    estimate.set_target(target)
    # check the estimates is correct
    assert_allclose(affine.h_matrix, estimate.h_matrix, atol=1e-3, rtol=1e-3)


def test_align_2d_affine_as_non_alignment():
    linear_component = np.array([[1, -6],
                                 [-3, 2]])
    translation_component = np.array([7, -8])
    h_matrix = np.eye(3, 3)
    h_matrix[:-1, :-1] = linear_component
    h_matrix[:-1, -1] = translation_component
    affine = Affine(h_matrix)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = affine.apply(source)
    # estimate the transform from source and source
    estimate = AlignmentAffine(source, source)
    # and set the h_matrix
    non_align = estimate.as_non_alignment()
    # check the estimates is correct
    assert_allclose(non_align.h_matrix, estimate.h_matrix)
    assert(type(non_align) == Affine)


# TODO check from_vector, from_vector_inplace works correctly


# SIMILARITY

def test_align_2d_similarity():
    linear_component = np.array([[2, -6],
                                 [6, 2]])
    translation_component = np.array([7, -8])
    h_matrix = np.eye(3, 3)
    h_matrix[:-1, :-1] = linear_component
    h_matrix[:-1, -1] = translation_component
    similarity = Similarity(h_matrix)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = similarity.apply(source)
    # estimate the transform from source and target
    estimate = AlignmentSimilarity(source, target)
    # check the estimates is correct
    assert_allclose(similarity.h_matrix,
                    estimate.h_matrix)


def test_align_2d_similarity_set_target():
    linear_component = np.array([[2, -6],
                                 [6, 2]])
    translation_component = np.array([7, -8])
    h_matrix = np.eye(3, 3)
    h_matrix[:-1, :-1] = linear_component
    h_matrix[:-1, -1] = translation_component
    similarity = Similarity(h_matrix)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = similarity.apply(source)
    # estimate the transform from source to source
    estimate = AlignmentSimilarity(source, source, allow_mirror=True)
    # and set the target
    estimate.set_target(target)
    # check the estimates is correct
    assert_allclose(similarity.h_matrix,
                    estimate.h_matrix)


# ROTATION

def test_align_2d_rotation():
    rotation_matrix = np.array([[0, 1],
                                [-1, 0]])
    rotation = Rotation(rotation_matrix)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = rotation.apply(source)
    # estimate the transform from source and target
    estimate = AlignmentRotation(source, target)
    # check the estimates is correct
    assert_allclose(rotation.h_matrix,
                    estimate.h_matrix, atol=1e-14)


def test_align_2d_rotation_allow_mirror():
    s_init = PointCloud(np.array([[-1., 1.], [1., 1.], [1., -1.], [-1., -1.]]))
    s_trg = PointCloud(np.array([[1., -1.], [1., 1.], [-1., 1.], [-1., -1.]]))
    # estimate the transform from source and target with mirroring allowed
    tr = AlignmentRotation(s_init, s_trg, allow_mirror=True)
    s_final = tr.apply(s_init)
    assert_allclose(s_final.points, s_trg.points, atol=1e-14)
    # estimate the transform from source and target with mirroring allowed
    tr = AlignmentRotation(s_init, s_trg, allow_mirror=False)
    s_final = tr.apply(s_init)
    assert_allclose(s_final.points, np.array([[-1., -1.], [-1., 1.], [1., 1.],
                                              [1., -1.]]), atol=1e-14)


def test_align_2d_rotation_set_target():
    rotation_matrix = np.array([[0, 1],
                                [-1, 0]])
    rotation = Rotation(rotation_matrix)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = rotation.apply(source)
    # estimate the transform from source and source
    estimate = AlignmentRotation(source, source)
    # and set the target
    estimate.set_target(target)
    # check the estimates is correct
    assert_allclose(rotation.h_matrix,
                    estimate.h_matrix, atol=1e-14)


def test_align_2d_rotation_set_rotation_matrix():
    rotation_matrix = np.array([[0, 1],
                                [-1, 0]])
    rotation = Rotation(rotation_matrix)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = rotation.apply(source)
    # estimate the transform from source and source
    estimate = AlignmentRotation(source, source)
    # and set the target
    estimate.set_rotation_matrix(rotation.rotation_matrix)
    # check the estimates is correct
    assert_allclose(target.points,
                    estimate.target.points, atol=1e-14)


# UNIFORM SCALE

def test_align_2d_uniform_scale():
    scale = UniformScale(2.5, 2)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = scale.apply(source)
    # estimate the transform from source and target
    estimate = AlignmentUniformScale(source, target)
    # check the estimates is correct
    assert_allclose(scale.h_matrix, estimate.h_matrix)


def test_align_2d_uniform_scale_set_target():
    scale = UniformScale(2.5, 2)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = scale.apply(source)
    # estimate the transform from source and source
    estimate = AlignmentUniformScale(source, source)
    # and set the target
    estimate.set_target(target)
    # check the estimates is correct
    assert_allclose(scale.h_matrix, estimate.h_matrix)


# TRANSLATION

def test_align_2d_translation():
    t_vec = np.array([1, 2])
    translation = Translation(t_vec)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = translation.apply(source)
    # estimate the transform from source and target
    estimate = AlignmentTranslation(source, target)
    # check the estimates is correct
    assert_allclose(translation.h_matrix,
                    estimate.h_matrix)


def test_align_2d_translation_set_target():
    t_vec = np.array([1, 2])
    translation = Translation(t_vec)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = translation.apply(source)
    # estimate the transform from source to source..
    estimate = AlignmentTranslation(source, source)
    # and change the target.
    estimate.set_target(target)
    # check the estimates is correct
    assert_allclose(translation.h_matrix,
                    estimate.h_matrix)


def test_align_2d_translation_from_vector_inplace():
    t_vec = np.array([1, 2])
    translation = Translation(t_vec)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = translation.apply(source)
    # estimate the transform from source to source..
    estimate = AlignmentTranslation(source, source)
    # and update from_vector
    estimate._from_vector_inplace(t_vec)
    # check the estimates is correct
    assert_allclose(target.points,
                    estimate.target.points)


def test_align_2d_translation_from_vector():
    t_vec = np.array([1, 2])
    translation = Translation(t_vec)
    source = PointCloud(np.array([[0, 1],
                                  [1, 1],
                                  [-1, -5],
                                  [3, -5]]))
    target = translation.apply(source)
    # estimate the transform from source to source..
    estimate = AlignmentTranslation(source, source)
    # and update from_vector
    new_est = estimate.from_vector(t_vec)
    # check the original is unchanged
    assert_allclose(estimate.source.points, source.points)
    assert_allclose(estimate.target.points, source.points)
    # check the new estimate has the source and target correct
    assert_allclose(new_est.source.points, source.points)
    assert_allclose(new_est.target.points, target.points)



from menpo.image import Image, MaskedImage
import numpy as np


def test_rescale_pixels():
    img = Image.init_blank((10, 10), n_channels=1)
    img.pixels[:, 6:, 6:] = 2

    img_rescaled = img.rescale_pixels(0, 255)
    assert np.min(img_rescaled.pixels) == 0
    assert np.max(img_rescaled.pixels) == 255
    assert np.all(img_rescaled.pixels[:, 6:, 6:] == 255)


def test_rescale_pixels_all_channels():
    img = Image.init_blank((10, 10), n_channels=2)
    img.pixels[0, 6:, 6:] = 2
    img.pixels[1, 6:, 6:] = 4

    img_rescaled = img.rescale_pixels(0, 100, per_channel=False)
    assert np.min(img_rescaled.pixels) == 0
    assert np.max(img_rescaled.pixels) == 100
    assert np.all(img_rescaled.pixels[0, 6:, 6:] == 50)
    assert np.all(img_rescaled.pixels[1, 6:, 6:] == 100)


def test_rescale_pixels_per_channel():
    img = Image.init_blank((10, 10), n_channels=2)
    img.pixels[0, 6:, 6:] = 2
    img.pixels[1, 6:, 6:] = 4

    img_rescaled = img.rescale_pixels(0, 100, per_channel=True)
    assert np.min(img_rescaled.pixels) == 0
    assert np.max(img_rescaled.pixels) == 100
    assert np.all(img_rescaled.pixels[0, 6:, 6:] == 100)
    assert np.all(img_rescaled.pixels[1, 6:, 6:] == 100)


def test_rescale_pixels_only_masked():
    img = MaskedImage.init_blank((10, 10), n_channels=1, fill=1)
    img.pixels[0, 0, 0] = 0
    img.pixels[0, 6:, 6:] = 2
    img.mask.pixels[:, 6:, 6:] = False

    img_rescaled = img.rescale_pixels(0, 100)
    assert np.min(img_rescaled.pixels) == 0
    assert np.max(img_rescaled.pixels) == 100
    assert img_rescaled.pixels[0, 0, 0] == 0
    assert img_rescaled.pixels[0, 1, 1] == 100
    assert np.all(img_rescaled.mask.pixels == img.mask.pixels)



from .base import labeller
from .human import *
from .car import (
    car_streetscene_20_to_car_streetscene_view_0_8,
    car_streetscene_20_to_car_streetscene_view_1_14,
    car_streetscene_20_to_car_streetscene_view_2_10,
    car_streetscene_20_to_car_streetscene_view_3_14,
    car_streetscene_20_to_car_streetscene_view_4_14,
    car_streetscene_20_to_car_streetscene_view_5_10,
    car_streetscene_20_to_car_streetscene_view_6_14,
    car_streetscene_20_to_car_streetscene_view_7_8)
from .bounding_box import (bounding_box_to_bounding_box,
                           bounding_box_mirrored_to_bounding_box)
